# -*- coding: utf-8 -*-
"""
:mod:`orion.core.io.orion_cmdline_parser` -- Parse command line arguments for Orion
===================================================================================

.. module:: orion_cmdline_parser
   :platform: Unix
   :synopsis: Parsing and building of the command line input for using script

Simplify the parsing of a command line by storing every values inside an `OrderedDict`
mapping the name of the argument to its value as a key-pair relation. Positional arguments
are stored in the format `_X` where `X` represent the index of that argument inside
the command line.

CmdlineParser provides an interface to parse command line arguments from input but also
utility functions to build it again as a list or an already formatted string.
"""

from collections import defaultdict, OrderedDict
import copy
import errno
import logging
import os
import re
import shutil

from orion.core.io.cmdline_parser import CmdlineParser
from orion.core.io.convert import infer_converter_from_file_type


log = logging.getLogger(__name__)


def _is_nonprior_wave(arg):
    return arg.startswith('/') or arg == ''


class OrionCmdlineParser():
    """Python interface commandline parser

    The `OrionCmdlineParser` is used as a way to obtain the parsing process of Orion
    through a Python interface. It provides different dictionaries containing information
    for different parts of the parsing process. It also exposes methods to retrieve the
    command line with formatted values for parameters.

    Parameters
    ----------
    config_prefix : str, optional
        Prefix for the configuration file used by the parser to identify it.
    allow_non_existing_files : bool, optional
        If True, will parse all commandline but ignore non existing user script or configuration
        files if non existant. Default is False

    Attributes
    ----------
    parser : CmdlineParser
        Parser that will be used to parse the commandline.
    cmd_priors : OrderedDict
        An OrderedDict containing only the priors inside the commandline.
    file_priors : OrderedDict
        An OrderedDict obtained by parsing the config file, if one was found.
    priors : OrderedDict
        An OrderedDict obtained from merging `cmd_priors` and `file_priors`.
    user_script : str
        File path of the script executed (inferred from parsed commandline)
    config_prefix : str
        Prefix for the configuration file used by the parser to identify it.
    file_config_path : str
        If a config file was found, this file contain the path. `None` otherwise.
    converter : BaseConverter
        A BaseConverter object to parse the config file.

    Methods
    -------
    parse(commandline)
        Parses the commandline and populate the OrderedDict
    format(trial, experiment) : str
        Return the commandline with replaced values for priors and attributes

    """

    def __init__(self, config_prefix='config', allow_non_existing_files=False):
        """Create an `OrionCmdlineParser`."""
        self.parser = CmdlineParser()
        self.cmd_priors = OrderedDict()
        self.file_priors = OrderedDict()
        self.config_file_data = {}

        self.config_prefix = config_prefix
        self.file_config_path = None
        self.converter = None

        self.allow_non_existing_files = allow_non_existing_files
        self.user_script = None

        # Extraction methods for the file parsing part.
        self._extraction_method = {dict: self._extract_dict,
                                   defaultdict: self._extract_defaultdict,
                                   list: self._extract_list,
                                   str: self._extract_file_string}

        # Look for anything followed by a tilt and possible branching attributes + prior
        self.prior_regex = re.compile(r'(.+)~([\+\-\>]?.+)')

    def get_state_dict(self):
        """Give state dict that can be used to reconstruct the parser"""
        return dict(
            parser=self.parser.get_state_dict(),
            cmd_priors=list(map(list, self.cmd_priors.items())),
            file_priors=list(map(list, self.file_priors.items())),
            config_file_data=self.config_file_data,
            config_prefix=self.config_prefix,
            file_config_path=self.file_config_path,
            converter=self.converter.get_state_dict() if self.converter else None)

    def set_state_dict(self, state):
        """Reset the parser based on previous state"""
        self.parser.set_state_dict(state['parser'])

        self.cmd_priors = OrderedDict(state['cmd_priors'])
        self.file_priors = OrderedDict(state['file_priors'])

        self.config_file_data = state['config_file_data']
        self.config_prefix = state['config_prefix']
        self.file_config_path = state['file_config_path']

        if self.file_config_path:
            self.converter = infer_converter_from_file_type(self.file_config_path)
            self.converter.set_state_dict(state['converter'])

    def parse(self, commandline):
        """Parse the commandline given for the definition of priors.

        Parse the commandline for priors and check if a specific key is found to parse
        an additional configuration file. Then the definition of the priors are stored
        inside the `priors` attribute.

        Raises
        ------
        ValueError
            If a prior inside the commandline and the config file have the same name.

        """
        self.infer_user_script(commandline)
        replaced = self._replace_priors(commandline)
        configuration = self.parser.parse(replaced)
        self._build_priors(configuration)

        duplicated_priors = set(self.cmd_priors.keys()) & set(self.file_priors.keys())
        if duplicated_priors:
            raise ValueError("Conflict: definition of same prior in commandline and config: "
                             "{}".format(duplicated_priors))

    def infer_user_script(self, user_args):
        """Infer the script name and perform some checks"""
        if not user_args:
            return

        # TODO: Parse commandline for any options to python and pick the script filepath properly
        if user_args[0] == 'python':
            user_script = user_args[1]
        else:
            user_script = user_args[0]

        if (not os.path.exists(user_script) and not shutil.which(user_script) and
                not self.allow_non_existing_files):
            raise OSError(errno.ENOENT, "The path specified for the script does not exist",
                          user_script)

        self.user_script = user_script

    @property
    def priors(self):
        """Return an OrderedDict obtained from merging `cmd_priors` and `file_priors`."""
        priors = copy.deepcopy(self.file_priors)
        priors.update(self.cmd_priors)
        return priors

    @staticmethod
    def _replace_priors(args):
        """Change directly name priors to more general form.

        Pass through the current commandline arguments and replace priors of the form
        `--<name>~<prior>` to the more general one `--<name> orion~<prior>` to stay consistent
        with parameters passed as list.

        Parameters
        ----------
        args: list
            A list of the current commandline arguments.

        Returns
        -------
        list
            A list composed of the same elements as `args` augmented with the new form
            of the priors.

        """
        replaced = []
        for item in args:
            if item.startswith('-'):
                # Get the prior part after the `~`
                parts = item.split('~')

                if len(parts) > 1 and _is_nonprior_wave(parts[1]):
                    replaced.append(item)
                    continue

                # If the argument was defined has a long one but only has a single letter
                # then it needs to be shortened.
                if parts[0].startswith('--') and len(parts[0]) == 3:
                    parts[0] = parts[0][1:]

                replaced.append(parts[0])

                if len(parts) > 1:
                    replaced.append('orion~' + parts[1])
            else:
                replaced.append(item)

        return replaced

    def _build_priors(self, configuration):
        """Create OrderedDict from priors only.

        Loop through every commandline arguments and check if it might correspond to a prior or a
        configuration file. Configuration file is parsed with `_load_config` while cmdline priors
        are extracted with `_extract_prior`.

        Parameters
        ----------
        configuration: OrderedDict
            The original configuration from which to extract OrderedDict.

        """
        for key, value in configuration.items():
            if key == self.config_prefix:
                self.file_config_path = value
                self._load_config(value)
            else:
                self._extract_prior(key, value, self.cmd_priors)

    def _load_config(self, path):
        """Load configuration file.

        Load the configuration file associated with the `config` key. Will try to resolve a
        valid converter for the file extension (yaml, json or other). Content will be put
        inside the `self.file_priors` attribute. Once the data has been parsed from the file,
        the corresponding configuration will be created.

        Parameters
        ----------
        path: string
            Path to the configuration file.

        """
        if not os.path.exists(path):
            if self.allow_non_existing_files:
                log.info("The path specified for the script config does not exist: %s", path)
                return
            else:
                raise OSError(
                    errno.ENOENT,
                    "The path specified for the script config does not exist",
                    path)

        self.converter = infer_converter_from_file_type(path)
        self.config_file_data = self.converter.parse(path)
        self._extraction_method[type(self.config_file_data)]("", self.config_file_data)

    def _extract_defaultdict(self, current_depth, ex_dict):
        for key, value in ex_dict.items():
            sub_depth = current_depth + '/' + str(key)

            try:
                if isinstance(value, str):
                    value = 'orion~' + value
                self._extraction_method[type(value)](sub_depth, value)
            except KeyError:
                pass

    def _extract_dict(self, current_depth, ex_dict):
        """Recursively extract data from dictionary.

        Loop through the pairs of key/value inside `ex_dict` to find information regarding
        priors defined inside the config file.

        Parameters
        ----------
        current_depth: string
            String corresponding to the namespace at the current depth.
        ex_dict: dict
            Dictionary to loop through.

        """
        for key, value in ex_dict.items():
            sub_depth = current_depth + '/' + str(key)

            try:
                self._extraction_method[type(value)](sub_depth, value)
            except KeyError:
                pass

    def _extract_list(self, current_depth, ex_list):
        """Recursively extract data from list.

        Loop through the values inside `ex_list` to find information regarding priors defined
        inside the config file.

        Parameters
        ----------
        current_depth: string
            String corresponding to the namespace at the current depth.
        ex_list: list
            List to loop through.

        """
        for i, value in enumerate(ex_list):
            sub_depth = current_depth + '/' + str(i)

            try:
                self._extraction_method[type(value)](sub_depth, value)
            except KeyError:
                pass

    def _extract_file_string(self, current_depth, value):
        """Extract the prior from a string

        This is used alongside `_extract_list` and `_extract_dict` to iterate through
        a config file and extract the prior corresponding to a string.

        Parameters
        ----------
        current_depth: string
            String corresponding to the namespace at the current depth.
        value: string
            Value from which to extract a prior.

        """
        substrings = value.split('~')

        if len(substrings) == 1:
            return

        self._extract_prior(current_depth, value, self.file_priors)

    def _extract_prior(self, key, value, insert_into):
        """Insert parameters if it has a prior.

        Match the regex for priors with the `value` argument to extract the information
        regarding the prior. If it posseses such information, insert the parameters inside
        the `cmd_priors` attribute.

        Parameters
        ----------
        key: str
            Current key for the element inside the `OrderedDict` of commandline arguments.
        Will correspond to `orion` if it is related to priors.

        value:
            Possible parameter to parse through the regex.

        insert_into: OrderedDict
            Collections into which to insert the current prior.

        """
        if not isinstance(value, str):
            return

        prior = self.prior_regex.match(value)
        if prior is None:
            return

        # Skip first group because it will always correspond to `orion`.
        _, expression = prior.groups(2)

        name = key
        if not name.startswith('/'):
            name = '/' + name

        insert_into[name] = expression

    def _should_not_be_built(self, expression):
        """Check if an expression should be built or not.

        When parsing priors, we might encounter ones that use the conflicts solving notation.
        Some of these tokens need to be removed (like the `+` sign) so that the prior can be built
        whilst other must not be built because they do not add another dimension to the Space.
        Such tokens are: `-` and `>`.

        Parameters
        ----------
        expression: str
            The expression to be evaluated.

        """
        for token in self._invalid_priors_tokens:
            if expression.startswith(token):
                return True

        return False

    def format(self, config_path=None, trial=None, experiment=None):
        """Create the commandline for the user's script.

        Recreate the commandline passed to Orion for the user's script by replacing the instances
        of priors' expression by their actual value inside the `trial`. If a `config_path` is given,
        use the config file template to generate a new temporary one. Any templated argument of the
        form `{trial.xxx}` or `{exp.xxx}` will be replaced by their corresponding value.

        Parameters
        ----------
        config_path: str
            Path to the temporary config file. Must be given if the parser has a `file_config_path`.
        trial: `orion.core.worker.trial.Trial`
            A `Trial` object containing the values for the priors.
        experiment: `orion.core.worker.experiment.Experiment`
            An `Experiment` object containing information relative to the `trial`'s experiment.

        Returns
        -------
        list
            The commandline arguments.

        Raises
        ------
        ValueError
            If the configuration contains a config file but `format()` is called without the
            argument `config_path`.

        """
        if self.file_config_path and config_path is None:
            raise ValueError('The configuration contains a config file. '
                             'Cannot format without a `config_path` argument.')
        elif self.file_config_path:
            self._create_config_file(config_path, trial)
        configuration = self._build_configuration(trial)

        if config_path is not None:
            configuration[self.config_prefix] = config_path

        templated = self.parser.format(configuration)

        trial_and_exp = dict(trial=trial, exp=experiment)

        for idx, item in enumerate(templated):
            templated[idx] = item.format(**trial_and_exp)

        return templated

    def _create_config_file(self, config_path, trial):
        # Create a copy of the template
        instance = copy.deepcopy(self.config_file_data)

        for name, value in trial.params.items():
            # The param will only correspond to config keyd
            # that require a prior, so we make sure to skip
            # the ones that do not.
            if name not in self.file_priors.keys():
                continue

            # Since namespace start with '/', we must skip
            # the first element of the list.
            path = name.split('/')[1:]
            current_depth = instance

            for key in path:
                # If we meet a list, the key might correspond
                # to the index of a dictionary in that list
                if isinstance(current_depth, list):
                    if not key.isdigit():
                        continue

                    key = int(key)

                    # Make sure the key is not out of bound
                    try:
                        current_depth[key]
                    except IndexError:
                        break

                if isinstance(current_depth[key], str):
                    current_depth[key] = value
                else:
                    current_depth = current_depth[key]

        self.converter.generate(config_path, instance)

    def _build_configuration(self, trial):
        configuration = copy.deepcopy(self.parser.arguments)

        for name, value in trial.params.items():
            name = name.lstrip('/')
            configuration[name] = value

        return configuration

    def priors_to_normal(self):
        """Remove the namespace `/` prefix from priors."""
        return {key.lstrip('/'): arg for key, arg in self.cmd_priors.items()}



# -*- coding: utf-8 -*-
"""
:mod:`orion.core.utils.flatten` -- Flatten and unflatten dicts
==============================================================

.. module:: flatten
   :platform: Unix
   :synopsis: Turn deep dictionaries into flat key.subkey versions and vice-versa.

"""


import copy


def flatten(dictionary):
    """Turn all nested dict keys into a {key}.{subkey} format"""
    def _flatten(dictionary):
        if dictionary == {}:
            return dictionary

        key, value = dictionary.popitem()
        if not isinstance(value, dict) or not value:
            new_dictionary = {key: value}
            new_dictionary.update(flatten(dictionary))
            return new_dictionary

        flat_sub_dictionary = flatten(value)
        for flat_sub_key in list(flat_sub_dictionary.keys()):
            flat_key = key + '.' + flat_sub_key
            flat_sub_dictionary[flat_key] = flat_sub_dictionary.pop(flat_sub_key)

        new_dictionary = flat_sub_dictionary
        new_dictionary.update(flatten(dictionary))
        return new_dictionary

    return _flatten(copy.deepcopy(dictionary))


def unflatten(dictionary):
    """Turn all keys with format {key}.{subkey} into nested dictionaries"""
    unflattened_dictionary = dict()
    for key, value in dictionary.items():
        parts = key.split(".")
        sub_dictionary = unflattened_dictionary
        for part in parts[:-1]:
            if part not in sub_dictionary:
                sub_dictionary[part] = dict()
            sub_dictionary = sub_dictionary[part]
        sub_dictionary[parts[-1]] = value
    return unflattened_dictionary



# -*- coding: utf-8 -*-
"""
:mod:`orion.core.utils.exception` -- Custom exceptions for Oríon
================================================================

.. module:: exception
   :platform: Unix
   :synopsis: Custom exceptions.
"""


NO_CONFIGURATION_FOUND = """\
No commandline configuration found for new experiment."""


NO_EXP_NAME_PROVIDED = """\
No name provided for the experiment."""


class NoConfigurationError(Exception):
    """Raise when commandline configuration is empty."""

    def __init__(self, message=NO_CONFIGURATION_FOUND):
        super().__init__(message)


class NoNameError(Exception):
    """Raise when no name is provided for an experiment."""

    def __init__(self, message=NO_EXP_NAME_PROVIDED):
        super().__init__(message)


class CheckError(Exception):
    """Raise when a check has failed."""

    pass


class RaceCondition(Exception):
    """Raise when a race condition occured."""

    pass


MISSING_RESULT_FILE = """
Cannot parse result file.

Make sure to report results in file `$ORION_RESULTS_PATH`.
This can be done with `orion.client.cli.report_objective()`.
"""


class MissingResultFile(Exception):
    """Raise when no result file (or empty) at end of trial execution."""

    def __init__(self, message=MISSING_RESULT_FILE):
        super().__init__(message)


BRANCHING_ERROR_MESSAGE = """\
Configuration is different and generates a branching event:
{}

Hint
----

This error is typically caused by the following 2 reasons:
  1) Commandline calls where arguments are different from one worker to another
     (think of paths that are worker specific). There will be --cli-change-type
     in the error message above if it is the case.
  2) User script that writes to the repository of the script, causing changes in the code
     and therefore leading to branching events. There will be --code-change-type
     in the error message above if it is the case.

For each case you should:
  1) Use --non-monitored-arguments [ARGUMENT_NAME]
     (where you argument would be --argument-name, note the lack of dashes at
      the beginning and the underscores instead of dashes between words)
     The commandline argument only support one entry. To ignore many arguments,
     you can use the option in a local config file, or in the global config file:
     ```
     evc:
         non_monitored_arguments: ['FIRST_ARG', 'ANOTHER_ARG']
     ```

  2) Avoid writing data in your repository. It should only be code anyway, right? :)
     Otherwise, you can ignore code changes altogether with option --ignore-code-changes.

"""


class BranchingEvent(Exception):
    """Raise when conflicts could not be automatically resolved."""

    def __init__(self, status, message=BRANCHING_ERROR_MESSAGE):
        super().__init__(message.format(status))


class SampleTimeout(Exception):
    """Raised when the algorithm is not able to sample new unique points in time"""

    pass


class WaitingForTrials(Exception):
    """Raised when the algorithm needs to wait for some trials to complete before it can suggest new
    ones
    """

    pass


class BrokenExperiment(Exception):
    """Raised when too many trials failed in an experiment and it is now considered broken"""

    pass


INEXECUTABLE_USER_SCRIPT_ERROR_MESSAGE = """\
User script is not executable:
{}

Use chmod +x <your script> to make it executable.
If your script is written in python, you can otherwise
use `python <your script>` instead of `./<your script>`.

"""


class InexecutableUserScript(PermissionError):
    """Raised during consumption when user script is not executable"""

    def __init__(self, cmdline, message=INEXECUTABLE_USER_SCRIPT_ERROR_MESSAGE):
        super().__init__(message.format(cmdline))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
:mod:`orion.core.cli` -- Functions that define console scripts
================================================================

.. module:: cli
   :platform: Unix
   :synopsis: Helper functions to setup an experiment and execute it.

"""
import logging

from orion.core.cli.base import OrionArgsParser
from orion.core.utils import module_import

log = logging.getLogger(__name__)


def load_modules_parser(orion_parser):
    """Search through the `cli` folder for any module containing a `get_parser` function"""
    modules = module_import.load_modules_in_path('orion.core.cli',
                                                 lambda m: hasattr(m, 'add_subparser'))
    for module in modules:
        get_parser = getattr(module, 'add_subparser')
        get_parser(orion_parser.get_subparsers())


def main(argv=None):
    """Entry point for `orion.core` functionality."""
    # Fetch experiment name, user's script path and command line arguments
    # Use `-h` option to show help

    orion_parser = OrionArgsParser()

    load_modules_parser(orion_parser)

    return orion_parser.execute(argv)


if __name__ == "__main__":
    returncode = main()
    if returncode > 0:
        raise SystemExit(returncode)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Collection of tests for :mod:`orion.core.worker.consumer`."""
import datetime
import time

import pytest

import orion.core.io.experiment_builder as experiment_builder
from orion.core.utils.format_trials import tuple_to_trial
from orion.core.worker.trial_pacemaker import TrialPacemaker
from orion.storage.base import get_storage


@pytest.fixture
def config(exp_config):
    """Return a configuration."""
    config = exp_config[0][0]
    config['space'] = {'x': 'uniform(-50, 50)'}
    config['name'] = 'exp'
    return config


@pytest.fixture
def exp(config):
    """Return an Experiment."""
    return experiment_builder.build(**config)


@pytest.fixture
def trial(exp):
    """Return a Trial which is registered in DB."""
    trial = tuple_to_trial((1.0,), exp.space)
    heartbeat = datetime.datetime.utcnow()
    trial.experiment = exp.id
    trial.status = 'reserved'
    trial.heartbeat = heartbeat

    get_storage().register_trial(trial)

    return trial


@pytest.mark.usefixtures("create_db_instance")
def test_trial_update_heartbeat(exp, trial):
    """Test that the heartbeat of a trial has been updated."""
    trial_monitor = TrialPacemaker(trial, wait_time=1)

    trial_monitor.start()
    time.sleep(2)

    trials = exp.fetch_trials_by_status('reserved')

    assert trial.heartbeat != trials[0].heartbeat

    heartbeat = trials[0].heartbeat

    time.sleep(2)

    trials = exp.fetch_trials_by_status(status='reserved')

    assert heartbeat != trials[0].heartbeat
    trial_monitor.stop()


@pytest.mark.usefixtures("create_db_instance")
def test_trial_heartbeat_not_updated(exp, trial):
    """Test that the heartbeat of a trial is not updated when trial is not longer reserved."""
    trial_monitor = TrialPacemaker(trial, wait_time=1)

    trial_monitor.start()
    time.sleep(2)

    trials = exp.fetch_trials_by_status('reserved')

    assert trial.heartbeat != trials[0].heartbeat

    get_storage().set_trial_status(trial, status='interrupted')

    time.sleep(2)

    # `join` blocks until all thread have finish executing. So, the test will hang if it fails.
    trial_monitor.join()
    assert 1


@pytest.mark.usefixtures("create_db_instance")
def test_trial_heartbeat_not_updated_inbetween(exp, trial):
    """Test that the heartbeat of a trial is not updated before wait time."""
    trial_monitor = TrialPacemaker(trial, wait_time=5)

    trial_monitor.start()
    time.sleep(1)

    trials = exp.fetch_trials_by_status('reserved')
    assert trial.heartbeat.replace(microsecond=0) == trials[0].heartbeat.replace(microsecond=0)

    heartbeat = trials[0].heartbeat

    time.sleep(6)

    trials = exp.fetch_trials_by_status(status='reserved')

    assert heartbeat != trials[0].heartbeat
    trial_monitor.stop()



# -*- coding: utf-8 -*-
"""
:mod:`orion.client.experiment` -- Experiment wrapper client
===========================================================

.. module:: experiment
   :platform: Unix
   :synopsis: Wraps the core Experiment object to provide further functionalities for the user

"""
import atexit
import functools
import logging
import sys

from numpy import inf as infinity

from orion.core.io.database import DuplicateKeyError
from orion.core.utils.exceptions import BrokenExperiment, SampleTimeout, WaitingForTrials
from orion.core.utils.flatten import flatten, unflatten
import orion.core.utils.format_trials as format_trials
import orion.core.worker
from orion.core.worker.trial import Trial
from orion.core.worker.trial_pacemaker import TrialPacemaker
from orion.plotting.base import PlotAccessor
from orion.storage.base import FailedUpdate


log = logging.getLogger(__name__)


def set_broken_trials(client):
    """Release all trials with status broken if the process exits without releasing them."""
    if sys.exc_info()[0] is KeyboardInterrupt:
        status = 'interrupted'
    else:
        status = 'broken'

    for trial_id in list(client._pacemakers.keys()):  # pylint: disable=protected-access
        trial = client.get_trial(uid=trial_id)
        if trial is None:
            log.warning('Trial {} was not found in storage, could not set status to `broken`.')
            continue
        client.release(trial, status=status)


# pylint: disable=too-many-public-methods
class ExperimentClient:
    """ExperimentClient providing all functionalities for the python API

    Note that the ExperimentClient is not meant to be instantiated by the user.
    Look at `orion.client.create_experiment` to build an ExperimentClient.

    Parameters
    ----------
    experiment: `orion.core.worker.experiment.Experiment`
        Experiment object serving for interaction with storage
    producer: `orion.core.worker.producer.Producer`
        Producer object used to produce new trials.

    """

    def __init__(self, experiment, producer, heartbeat=None):
        self._experiment = experiment
        self._producer = producer
        self._pacemakers = {}
        self.set_broken_trials = functools.partial(set_broken_trials, client=self)
        if heartbeat is None:
            heartbeat = orion.core.config.worker.heartbeat
        self.heartbeat = heartbeat
        self.plot = PlotAccessor(self)
        atexit.register(self.set_broken_trials)

    ###
    # Attributes
    ###

    @property
    def name(self):
        """Return the name of the experiment in the database."""
        return self._experiment.name

    # pylint: disable=invalid-name
    @property
    def id(self):
        """Return the id of the experiment in the database."""
        return self._experiment.id

    @property
    def version(self):
        """Version of the experiment."""
        return self._experiment.version

    @property
    def max_trials(self):
        """Max-trials to execute before stopping the experiment."""
        return self._experiment.max_trials

    @property
    def metadata(self):
        """Metadata of the experiment."""
        return self._experiment.metadata

    @property
    def space(self):
        """Return problem's parameter `orion.algo.space.Space`."""
        return self._experiment.space

    @property
    def algorithms(self):
        """Algorithms of the experiment."""
        return self._experiment.algorithms

    @property
    def is_done(self):
        """Return True, if this experiment is considered to be finished.

        1. Count how many trials have been completed and compare with `max_trials`.
        2. Ask `algorithms` if they consider there is a chance for further improvement.
        """
        return self._experiment.is_done

    @property
    def is_broken(self):
        """Return True, if this experiment is considered to be broken.

        Count how many trials are broken and return True if that number has reached
        as given threshold.
        """
        return self._experiment.is_broken

    @property
    def configuration(self):
        """Return a copy of an `Experiment` configuration as a dictionary."""
        return self._experiment.configuration

    @property
    def stats(self):
        """Calculate a stats dictionary for this particular experiment.

        Returns
        -------
        stats : dict

        Stats
        -----
        trials_completed : int
           Number of completed trials
        best_trials_id : int
           Unique identifier of the `Trial` object in the database which achieved
           the best known objective result.
        best_evaluation : float
           Evaluation score of the best trial
        start_time : `datetime.datetime`
           When Experiment was first dispatched and started running.
        finish_time : `datetime.datetime`
           When Experiment reached terminating condition and stopped running.
        duration : `datetime.timedelta`
           Elapsed time.

        """
        return self._experiment.stats

    @property
    def node(self):
        """Node of the experiment in the version control tree."""
        return self._experiment.node

    @property
    def working_dir(self):
        """Working directory of the experiment."""
        return self._experiment.working_dir

    @property
    def producer(self):
        """Return the producer configuration of the experiment."""
        return self._experiment.producer

    ###
    # Queries
    ###

    def fetch_trials(self, with_evc_tree=False):
        """Fetch all trials of the experiment

        Parameters
        ----------
        with_evc_tree: bool, optional
            Fetch all trials from the EVC tree.
            Default: False

        """
        return self._experiment.fetch_trials(with_evc_tree=with_evc_tree)

    def get_trial(self, trial=None, uid=None):
        """Fetch a single trial

        Parameters
        ----------
        trial: Trial, optional
           trial object to retrieve from the database

        uid: str, optional
            trial id used to retrieve the trial object

        Returns
        -------
        return none if the trial is not found,

        Raises
        ------
        UndefinedCall
            if both trial and uid are not set

        AssertionError
            if both trial and uid are provided and they do not match

        """
        return self._experiment.get_trial(trial, uid)

    def fetch_trials_by_status(self, status, with_evc_tree=False):
        """Fetch all trials with the given status

        Trials are sorted based on `Trial.submit_time`

        :return: list of `Trial` objects
        """
        return self._experiment.fetch_trials_by_status(status, with_evc_tree=with_evc_tree)

    def fetch_noncompleted_trials(self, with_evc_tree=False):
        """Fetch non-completed trials of this `Experiment` instance.

        Trials are sorted based on `Trial.submit_time`

        .. note::

            It will return all non-completed trials, including new, reserved, suspended,
            interrupted and broken ones.

        :return: list of non-completed `Trial` objects
        """
        return self._experiment.fetch_noncompleted_trials(with_evc_tree=with_evc_tree)

    ###
    # Actions
    ###

    # pylint: disable=unused-argument
    def insert(self, params, results=None, reserve=False):
        """Insert a new trial.

        Parameters
        ----------
        params: dict
            Parameters of the new trial to add to the database. These parameters
            must comply with the space definition otherwise a ValueError will be raised.
        results: list, optional
            Results to be set for the new trial. Results must have the format
            {name: <str>: type: <'objective', 'constraint' or 'gradient'>, value=<float>} otherwise
            a ValueError will be raised.
            Note that passing results will mark the trial as completed and therefore cannot be
            reserved. The returned trial will have status 'completed'.
            If the results are invalid, the trial will still be inserted but reservation will be
            released.
        reserve: bool, optional
            If reserve=True, the inserted trial will be reserved. `reserve` cannot be True if
            `results` are given.
            Defaults to False.

        Returns
        -------
        `orion.core.worker.trial.Trial`
            The trial inserted in storage. If `reserve=True` and no results are given, the returned
            trial will be in a `reserved` status.

        Raises
        ------
        `ValueError`
            - If results are given and reserve=True
            - If params have invalid format
            - If results have invalid format
        `orion.core.io.database.DuplicateKeyError`
            - If a trial with identical params already exist for the current experiment.

        """
        if results and reserve:
            raise ValueError(
                'Cannot observe a trial and reserve it. A trial with results has status '
                '`completed` and cannot be reserved.')
        trial = format_trials.dict_to_trial(params, self.space)
        try:
            self._experiment.register_trial(trial, status='reserved')
            self._maintain_reservation(trial)
        except DuplicateKeyError as e:
            message = 'A trial with params {} already exist for experiment {}-v{}'.format(
                params, self.name, self.version)
            raise DuplicateKeyError(message) from e

        if results:
            try:
                self.observe(trial, results)
            except ValueError:
                self._release_reservation(trial)
                raise

            return trial

        if not reserve:
            self.release(trial)

        return trial

    def reserve(self, trial):
        """Reserve a trial.

        Set a trial status to reserve to ensure that concurrent process cannot work on it.
        Trials can only be reserved with status 'new', 'interrupted' or 'suspended'.

        Parameters
        ----------
        trial: `orion.core.worker.trial.Trial`
            Trial to reserve.

        Raises
        ------
        `RuntimeError`
            If trial is reserved by another process
        `ValueError`
            If the trial does not exist in storage.

        Notes
        -----
        When reserved, a `TrialPacemaker` is started to update an heartbeat in storage. The
        frequency of the heartbeat is configurable at creation of experiment
        or with `orion.core.config.worker.heartbeat`.
        If the process terminates unexpectedly, the heartbeat will cease and remote processes
        may reset the status of the trial to 'interrupted' when the heartbeat has not been updated
        since twice the value of `heartbeat`.

        """
        if trial.status == 'reserved' and trial.id in self._pacemakers:
            log.warning('Trial %s is already reserved.', trial.id)
            return
        elif trial.status == 'reserved' and trial.id not in self._pacemakers:
            raise RuntimeError('Trial {} is already reserved by another process.'.format(trial.id))
        try:
            self._experiment.set_trial_status(trial, 'reserved', heartbeat=self.heartbeat)
        except FailedUpdate as e:
            if self.get_trial(trial) is None:
                raise ValueError('Trial {} does not exist in database.'.format(trial.id)) from e
            raise RuntimeError('Could not reserve trial {}.'.format(trial.id)) from e

        self._maintain_reservation(trial)

    def release(self, trial, status='interrupted'):
        """Release a trial.

        Release the reservation and stop the heartbeat.

        Parameters
        ----------
        trial: `orion.core.worker.trial.Trial`
            Trial to reserve.
        status: str, optional
            Set the trial to given status while releasing the reservation.
            Defaults to 'interrupted'.

        Raises
        ------
        `RuntimeError`
            If reservation of the trial has been lost prior to releasing it.
        `ValueError`
            If the trial does not exist in storage.

        """
        try:
            self._experiment.set_trial_status(trial, status)
        except FailedUpdate as e:
            if self.get_trial(trial) is None:
                raise ValueError('Trial {} does not exist in database.'.format(trial.id)) from e
            raise RuntimeError(
                'Reservation for trial {} has been lost before release.'.format(trial.id)) from e
        finally:
            self._release_reservation(trial)

    def suggest(self):
        """Suggest a trial to execute.

        If any trial is available (new or interrupted), it selects one and reserves it.
        Otherwise, the algorithm is used to generate a new trial that is registered in storage and
        reserved.

        Returns
        -------
        `orior.core.worker.trial.Trial` or None
            Reserved trial for execution. Will return None if experiment is done.
            of if the algorithm cannot suggest until other trials complete.

        Raises
        ------
        `WaitingForTrials`
            if the experiment is not completed and algorithm needs to wait for some
            trials to complete before it can suggest new trials.

        `BrokenExperiment`
            if too many trials failed to run and the experiment cannot continue.
            This is determined by ``max_broken`` in the configuration of the experiment.

        `SampleTimeout`
            if the algorithm of the experiment could not sample new unique points.

        """
        if self.is_broken:
            raise BrokenExperiment("Trials failed too many times")

        if self.is_done:
            return None

        try:
            trial = orion.core.worker.reserve_trial(self._experiment, self._producer)

        except WaitingForTrials as e:
            if self.is_broken:
                raise BrokenExperiment("Trials failed too many times") from e

            raise e

        except SampleTimeout as e:
            if self.is_broken:
                raise BrokenExperiment("Trials failed too many times") from e

            raise e

        if trial is not None:
            self._maintain_reservation(trial)

        return trial

    def observe(self, trial, results):
        """Observe trial results

        Parameters
        ----------
        trial: `orion.core.worker.trial.Trial`
            Reserved trial to observe.
        results: list
            Results to be set for the new trial. Results must have the format
            {name: <str>: type: <'objective', 'constraint' or 'gradient'>, value=<float>} otherwise
            a ValueError will be raised. If the results are invalid, the trial will not be released.

        Returns
        -------
        `orion.core.worker.trial.Trial`
            The trial inserted in storage. If `reserve=True` and no results are given, the returned
            trial will be in a `reserved` status.

        Raises
        ------
        `ValueError`
            - If results have invalid format
            - If the trial does not exist in storage.
        `RuntimeError`
            If reservation of the trial has been lost prior to releasing it.

        """
        trial.results += [Trial.Result(**result) for result in results]
        try:
            self._experiment.update_completed_trial(trial)
            self.release(trial, 'completed')
        except FailedUpdate as e:
            if self.get_trial(trial) is None:
                raise ValueError('Trial {} does not exist in database.'.format(trial.id)) from e

            self._release_reservation(trial)
            raise RuntimeError('Reservation for trial {} has been lost.'.format(trial.id)) from e

    def workon(self, fct, max_trials=infinity, **kwargs):
        """Optimize a given function

        Parameters
        ----------
        fct: callable
            Function to optimize. Must take arguments provided by trial.params. Additional constant
            parameter can be passed as ``**kwargs`` to `workon`. Function must return the final
            objective.
        max_trials: int, optional
            Maximum number of trials to execute within `workon`. If the experiment or algorithm
            reach status is_done before, the execution of `workon` terminates.
        **kwargs
            Constant argument to pass to `fct` in addition to trial.params. If values in kwargs are
            present in trial.params, the latter takes precedence.

        Raises
        ------
        `ValueError`
             If results returned by `fct` have invalid format

        """
        trials = 0
        kwargs = flatten(kwargs)
        while not self.is_done and trials < max_trials:
            trial = self.suggest()
            if trial is None:
                log.warning('Algorithm could not sample new points')
                return trials
            kwargs.update(flatten(trial.params))
            results = fct(**unflatten(kwargs))
            self.observe(trial, results=results)
            trials += 1

        return trials

    def close(self):
        """Verify that no reserved trials are remaining and unregister atexit()."""
        if self._pacemakers:
            raise RuntimeError("There is still reserved trials: {}\nRelease all trials before "
                               "closing the client, using "
                               "client.release(trial).".format(self._pacemakers.keys()))

        atexit.unregister(self.set_broken_trials)

    ###
    # Private
    ###

    def __repr__(self):
        """Represent the object as a string."""
        return "Experiment(name=%s, version=%s)" % (self.name, self.version)

    def _verify_reservation(self, trial):
        if trial.id not in self._pacemakers:
            raise RuntimeError(
                'Trial {} had no pacemakers. Was is reserved properly?'.format(trial.id))

        if self.get_trial(trial).status != 'reserved':
            self._release_reservation(trial)
            raise RuntimeError(
                'Reservation for trial {} has been lost.'.format(trial.id))

    def _maintain_reservation(self, trial):
        self._pacemakers[trial.id] = TrialPacemaker(trial)
        self._pacemakers[trial.id].start()

    def _release_reservation(self, trial):
        if trial.id not in self._pacemakers:
            raise RuntimeError(
                'Trial {} had no pacemakers. Was is reserved properly?'.format(trial.id))
        self._pacemakers.pop(trial.id).stop()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Perform a functional test of the status command."""
import os

import pytest

import orion.core.cli


def test_no_experiments(clean_db, monkeypatch, capsys):
    """Test status with no experiments."""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))
    orion.core.cli.main(['status'])

    captured = capsys.readouterr().out

    assert captured == "No experiment found\n"


def test_no_version_backward_compatible(clean_db, one_experiment_no_version, capsys):
    """Test status with no experiments."""
    orion.core.cli.main(['status'])

    captured = capsys.readouterr().out

    expected = """\
test_single_exp-no-version-v1
=============================
empty


"""
    assert captured == expected


def test_python_api(clean_db, with_experiment_using_python_api, capsys):
    """Test status with experiments built using python api."""
    orion.core.cli.main(['status'])

    captured = capsys.readouterr().out

    expected = """\
test_single_exp-v1
==================
empty


from-python-api-v1
==================
empty


"""
    assert captured == expected


def test_missing_conf_file(clean_db, with_experiment_missing_conf_file, capsys):
    """Test status can handle experiments when the user script config file is missing"""
    orion.core.cli.main(['status'])

    captured = capsys.readouterr().out

    expected = """\
test_single_exp-v1
==================
empty


"""
    assert captured == expected


def test_experiment_without_trials_wout_ac(clean_db, one_experiment, capsys):
    """Test status with only one experiment and no trials."""
    orion.core.cli.main(['status'])

    captured = capsys.readouterr().out

    expected = """\
test_single_exp-v1
==================
empty


"""
    assert captured == expected


def test_experiment_wout_success_wout_ac(clean_db, single_without_success, capsys):
    """Test status with only one experiment and no successful trial."""
    orion.core.cli.main(['status'])

    captured = capsys.readouterr().out

    expected = """\
test_single_exp-v1
==================
status         quantity
-----------  ----------
broken                1
interrupted           1
new                   1
reserved              1
suspended             1


"""
    assert captured == expected


def test_experiment_number_same_list_status(clean_db,
                                            single_without_success, capsys):
    """Test status and list command output the consistent number of experiments"""
    orion.core.cli.main(['status'])

    captured = capsys.readouterr().out

    expected = """\
test_single_exp-v1
==================
status         quantity
-----------  ----------
broken                1
interrupted           1
new                   1
reserved              1
suspended             1


"""
    assert captured == expected

    orion.core.cli.main(['list'])

    captured = capsys.readouterr().out

    assert captured == " test_single_exp-v1\n"


def test_experiment_w_trials_wout_ac(clean_db, single_with_trials, capsys):
    """Test status with only one experiment and all trials."""
    orion.core.cli.main(['status'])

    captured = capsys.readouterr().out

    expected = """\
test_single_exp-v1
==================
status         quantity    min obj
-----------  ----------  ---------
broken                1
completed             1          0
interrupted           1
new                   1
reserved              1
suspended             1


"""
    assert captured == expected


def test_two_unrelated_w_trials_wout_ac(clean_db, unrelated_with_trials, capsys):
    """Test two unrelated experiments, with all types of trials."""
    orion.core.cli.main(['status'])

    captured = capsys.readouterr().out

    expected = """\
test_double_exp-v1
==================
status         quantity
-----------  ----------
broken                1
completed             1
interrupted           1
new                   1
reserved              1
suspended             1


test_single_exp-v1
==================
status         quantity    min obj
-----------  ----------  ---------
broken                1
completed             1          0
interrupted           1
new                   1
reserved              1
suspended             1


"""

    assert captured == expected


def test_two_related_w_trials_wout_ac(clean_db, family_with_trials, capsys):
    """Test two related experiments, with all types of trials."""
    orion.core.cli.main(['status'])

    captured = capsys.readouterr().out

    expected = """\
test_double_exp-v1
==================
status         quantity
-----------  ----------
broken                1
completed             1
interrupted           1
new                   1
reserved              1
suspended             1


  test_double_exp_child-v1
  ========================
  status         quantity
  -----------  ----------
  broken                1
  completed             1
  interrupted           1
  new                   1
  reserved              1
  suspended             1


"""

    assert captured == expected


def test_three_unrelated_wout_ac(clean_db, three_experiments_with_trials, capsys):
    """Test three unrelated experiments with all types of trials."""
    orion.core.cli.main(['status'])

    captured = capsys.readouterr().out

    expected = """\
test_double_exp-v1
==================
status         quantity
-----------  ----------
broken                1
completed             1
interrupted           1
new                   1
reserved              1
suspended             1


  test_double_exp_child-v1
  ========================
  status         quantity
  -----------  ----------
  broken                1
  completed             1
  interrupted           1
  new                   1
  reserved              1
  suspended             1


test_single_exp-v1
==================
status         quantity    min obj
-----------  ----------  ---------
broken                1
completed             1          0
interrupted           1
new                   1
reserved              1
suspended             1


"""

    assert captured == expected


def test_three_related_wout_ac(clean_db, three_family_with_trials, capsys):
    """Test three related experiments with all types of trials."""
    orion.core.cli.main(['status'])

    captured = capsys.readouterr().out

    expected = """\
test_double_exp-v1
==================
status         quantity
-----------  ----------
broken                1
completed             1
interrupted           1
new                   1
reserved              1
suspended             1


  test_double_exp_child-v1
  ========================
  status         quantity
  -----------  ----------
  broken                1
  completed             1
  interrupted           1
  new                   1
  reserved              1
  suspended             1


  test_double_exp_child2-v1
  =========================
  status         quantity
  -----------  ----------
  broken                1
  completed             1
  interrupted           1
  new                   1
  reserved              1
  suspended             1


"""

    assert captured == expected


def test_three_related_branch_wout_ac(clean_db, three_family_branch_with_trials, capsys):
    """Test three related experiments with all types of trials."""
    orion.core.cli.main(['status'])

    captured = capsys.readouterr().out

    expected = """\
test_double_exp-v1
==================
status         quantity
-----------  ----------
broken                1
completed             1
interrupted           1
new                   1
reserved              1
suspended             1


  test_double_exp_child-v1
  ========================
  status         quantity
  -----------  ----------
  broken                1
  completed             1
  interrupted           1
  new                   1
  reserved              1
  suspended             1


    test_double_exp_grand_child-v1
    ==============================
    status         quantity
    -----------  ----------
    broken                1
    completed             1
    interrupted           1
    new                   1
    reserved              1
    suspended             1


"""

    assert captured == expected


def test_one_wout_trials_w_a_wout_c(clean_db, one_experiment, capsys):
    """Test experiments, without trials, with --all."""
    orion.core.cli.main(['status', '--all'])

    captured = capsys.readouterr().out

    expected = """\
test_single_exp-v1
==================
id     status    best objective
-----  --------  ----------------
empty


"""

    assert captured == expected


def test_one_w_trials_w_a_wout_c(clean_db, single_with_trials, capsys):
    """Test experiment, with all trials, with --all."""
    orion.core.cli.main(['status', '--all'])

    captured = capsys.readouterr().out

    expected = """\
test_single_exp-v1
==================
id                                status         min obj
--------------------------------  -----------  ---------
9f360d1b4eb2707f19dd619d0d898dd9  broken
47564e5e390348b9d1335d4013895eb4  completed            0
aefd38473f108016fd4842aa855732ff  interrupted
0695f63ecaf7d78f4b85d4cb344e0dc0  new
b0ea9850c09370215b45b81edd33c7d3  reserved
b49e902aebccce14e834d96e411f896e  suspended


"""

    assert captured == expected


def test_one_wout_success_w_a_wout_c(clean_db, single_without_success, capsys):
    """Test experiment, without success, with --all."""
    orion.core.cli.main(['status', '--all'])

    captured = capsys.readouterr().out

    expected = """\
test_single_exp-v1
==================
id                                status
--------------------------------  -----------
9f360d1b4eb2707f19dd619d0d898dd9  broken
aefd38473f108016fd4842aa855732ff  interrupted
0695f63ecaf7d78f4b85d4cb344e0dc0  new
b0ea9850c09370215b45b81edd33c7d3  reserved
b49e902aebccce14e834d96e411f896e  suspended


"""

    assert captured == expected


def test_two_unrelated_w_a_wout_c(clean_db, unrelated_with_trials, capsys):
    """Test two unrelated experiments with --all."""
    orion.core.cli.main(['status', '--all'])

    captured = capsys.readouterr().out

    expected = """\
test_double_exp-v1
==================
id                                status
--------------------------------  -----------
c2187f4954884c801e423d851aec9a0b  broken
e42cc22a15188d72df315b9eac79c9c0  completed
b849f69cc3a77f39382d7435d0d41b14  interrupted
7fbbd152f7ca2c064bf00441e311609d  new
667513aa2cb2244bee9c4f41c7ff1cea  reserved
557b9fdb9f96569dff7eb2de10d3946f  suspended


test_single_exp-v1
==================
id                                status         min obj
--------------------------------  -----------  ---------
9f360d1b4eb2707f19dd619d0d898dd9  broken
47564e5e390348b9d1335d4013895eb4  completed            0
aefd38473f108016fd4842aa855732ff  interrupted
0695f63ecaf7d78f4b85d4cb344e0dc0  new
b0ea9850c09370215b45b81edd33c7d3  reserved
b49e902aebccce14e834d96e411f896e  suspended


"""

    assert captured == expected


def test_two_related_w_a_wout_c(clean_db, family_with_trials, capsys):
    """Test two related experiments with --all."""
    orion.core.cli.main(['status', '--all'])

    captured = capsys.readouterr().out

    expected = """\
test_double_exp-v1
==================
id                                status
--------------------------------  -----------
c2187f4954884c801e423d851aec9a0b  broken
e42cc22a15188d72df315b9eac79c9c0  completed
b849f69cc3a77f39382d7435d0d41b14  interrupted
7fbbd152f7ca2c064bf00441e311609d  new
667513aa2cb2244bee9c4f41c7ff1cea  reserved
557b9fdb9f96569dff7eb2de10d3946f  suspended


  test_double_exp_child-v1
  ========================
  id                                status
  --------------------------------  -----------
  9bd1ebc475bcb9e077a9e81a7c954a65  broken
  3c1af2af2c8dc9862df2cef0a65d6e1f  completed
  614ec3fc127d52129bc9d66d9aeec36c  interrupted
  4487e7fc87c288d254f94dfa82cd79cc  new
  7877287c718d7844570003fd654f66ba  reserved
  ff997e666e20c5a8c1a816dde0b5e2e9  suspended


"""

    assert captured == expected


def test_three_unrelated_w_a_wout_c(clean_db, three_experiments_with_trials, capsys):
    """Test three unrelated experiments with --all."""
    orion.core.cli.main(['status', '--all'])

    captured = capsys.readouterr().out

    expected = """\
test_double_exp-v1
==================
id                                status
--------------------------------  -----------
c2187f4954884c801e423d851aec9a0b  broken
e42cc22a15188d72df315b9eac79c9c0  completed
b849f69cc3a77f39382d7435d0d41b14  interrupted
7fbbd152f7ca2c064bf00441e311609d  new
667513aa2cb2244bee9c4f41c7ff1cea  reserved
557b9fdb9f96569dff7eb2de10d3946f  suspended


  test_double_exp_child-v1
  ========================
  id                                status
  --------------------------------  -----------
  9bd1ebc475bcb9e077a9e81a7c954a65  broken
  3c1af2af2c8dc9862df2cef0a65d6e1f  completed
  614ec3fc127d52129bc9d66d9aeec36c  interrupted
  4487e7fc87c288d254f94dfa82cd79cc  new
  7877287c718d7844570003fd654f66ba  reserved
  ff997e666e20c5a8c1a816dde0b5e2e9  suspended


test_single_exp-v1
==================
id                                status         min obj
--------------------------------  -----------  ---------
9f360d1b4eb2707f19dd619d0d898dd9  broken
47564e5e390348b9d1335d4013895eb4  completed            0
aefd38473f108016fd4842aa855732ff  interrupted
0695f63ecaf7d78f4b85d4cb344e0dc0  new
b0ea9850c09370215b45b81edd33c7d3  reserved
b49e902aebccce14e834d96e411f896e  suspended


"""

    assert captured == expected


def test_three_related_w_a_wout_c(clean_db, three_family_with_trials, capsys):
    """Test three related experiments with --all."""
    orion.core.cli.main(['status', '--all'])

    captured = capsys.readouterr().out

    expected = """\
test_double_exp-v1
==================
id                                status
--------------------------------  -----------
c2187f4954884c801e423d851aec9a0b  broken
e42cc22a15188d72df315b9eac79c9c0  completed
b849f69cc3a77f39382d7435d0d41b14  interrupted
7fbbd152f7ca2c064bf00441e311609d  new
667513aa2cb2244bee9c4f41c7ff1cea  reserved
557b9fdb9f96569dff7eb2de10d3946f  suspended


  test_double_exp_child-v1
  ========================
  id                                status
  --------------------------------  -----------
  9bd1ebc475bcb9e077a9e81a7c954a65  broken
  3c1af2af2c8dc9862df2cef0a65d6e1f  completed
  614ec3fc127d52129bc9d66d9aeec36c  interrupted
  4487e7fc87c288d254f94dfa82cd79cc  new
  7877287c718d7844570003fd654f66ba  reserved
  ff997e666e20c5a8c1a816dde0b5e2e9  suspended


  test_double_exp_child2-v1
  =========================
  id                                status
  --------------------------------  -----------
  2c2b64df1859b45a0b01362ca146584a  broken
  225b4a17dd29d5c0423a81c1ddda8f0e  completed
  fb0bb45bd0a45225e2368a8158df0427  interrupted
  57bd3071c7c1c39ceb997a7b37c5470d  new
  673449a3910fdea777ac8cb8576cdbe3  reserved
  01a38cce74701c3b40eb3d92143bc90f  suspended


"""

    assert captured == expected


def test_three_related_branch_w_a_wout_c(clean_db, three_family_branch_with_trials, capsys):
    """Test three related experiments in a branch with --all."""
    orion.core.cli.main(['status', '--all'])

    captured = capsys.readouterr().out

    expected = """\
test_double_exp-v1
==================
id                                status
--------------------------------  -----------
c2187f4954884c801e423d851aec9a0b  broken
e42cc22a15188d72df315b9eac79c9c0  completed
b849f69cc3a77f39382d7435d0d41b14  interrupted
7fbbd152f7ca2c064bf00441e311609d  new
667513aa2cb2244bee9c4f41c7ff1cea  reserved
557b9fdb9f96569dff7eb2de10d3946f  suspended


  test_double_exp_child-v1
  ========================
  id                                status
  --------------------------------  -----------
  9bd1ebc475bcb9e077a9e81a7c954a65  broken
  3c1af2af2c8dc9862df2cef0a65d6e1f  completed
  614ec3fc127d52129bc9d66d9aeec36c  interrupted
  4487e7fc87c288d254f94dfa82cd79cc  new
  7877287c718d7844570003fd654f66ba  reserved
  ff997e666e20c5a8c1a816dde0b5e2e9  suspended


    test_double_exp_grand_child-v1
    ==============================
    id                                status
    --------------------------------  -----------
    82f82a325b7cf09251a34c9264e1812a  broken
    94baf74a4e94f800b6865d8ab5675428  completed
    e24b2e542c0869064abdb20c2de250eb  interrupted
    960bad983c3ee6349b8767fe452ecbb3  new
    8d8578e31c740c1c0fc385c961702481  reserved
    584375e2b32af0573f4692cb47a2ec99  suspended


"""

    assert captured == expected


def test_two_unrelated_w_c_wout_a(clean_db, unrelated_with_trials, capsys):
    """Test two unrelated experiments with --collapse."""
    orion.core.cli.main(['status', '--collapse'])

    captured = capsys.readouterr().out

    expected = """\
test_double_exp-v1
==================
status         quantity
-----------  ----------
broken                1
completed             1
interrupted           1
new                   1
reserved              1
suspended             1


test_single_exp-v1
==================
status         quantity    min obj
-----------  ----------  ---------
broken                1
completed             1          0
interrupted           1
new                   1
reserved              1
suspended             1


"""

    assert captured == expected


def test_two_related_w_c_wout_a(clean_db, family_with_trials, capsys):
    """Test two related experiments with --collapse."""
    orion.core.cli.main(['status', '--collapse'])

    captured = capsys.readouterr().out

    expected = """\
test_double_exp-v1
==================
status         quantity
-----------  ----------
broken                1
completed             1
interrupted           1
new                   2
reserved              1
suspended             1


"""

    assert captured == expected


def test_three_unrelated_w_c_wout_a(clean_db, three_experiments_with_trials, capsys):
    """Test three unrelated experiments with --collapse."""
    orion.core.cli.main(['status', '--collapse'])

    captured = capsys.readouterr().out

    expected = """\
test_double_exp-v1
==================
status         quantity
-----------  ----------
broken                1
completed             1
interrupted           1
new                   2
reserved              1
suspended             1


test_single_exp-v1
==================
status         quantity    min obj
-----------  ----------  ---------
broken                1
completed             1          0
interrupted           1
new                   1
reserved              1
suspended             1


"""

    assert captured == expected


def test_three_related_w_c_wout_a(clean_db, three_family_with_trials, capsys):
    """Test three related experiments with --collapse."""
    orion.core.cli.main(['status', '--collapse'])

    captured = capsys.readouterr().out

    expected = """\
test_double_exp-v1
==================
status         quantity
-----------  ----------
broken                1
completed             1
interrupted           1
new                   3
reserved              1
suspended             1


"""

    assert captured == expected


def test_three_related_branch_w_c_wout_a(clean_db, three_family_branch_with_trials, capsys):
    """Test three related experiments with --collapse."""
    orion.core.cli.main(['status', '--collapse'])

    captured = capsys.readouterr().out

    expected = """\
test_double_exp-v1
==================
status         quantity
-----------  ----------
broken                1
completed             1
interrupted           1
new                   3
reserved              1
suspended             1


"""

    assert captured == expected


def test_two_unrelated_w_ac(clean_db, unrelated_with_trials, capsys):
    """Test two unrelated experiments with --collapse and --all."""
    orion.core.cli.main(['status', '--collapse', '--all'])

    captured = capsys.readouterr().out

    expected = """\
test_double_exp-v1
==================
id                                status
--------------------------------  -----------
c2187f4954884c801e423d851aec9a0b  broken
e42cc22a15188d72df315b9eac79c9c0  completed
b849f69cc3a77f39382d7435d0d41b14  interrupted
7fbbd152f7ca2c064bf00441e311609d  new
667513aa2cb2244bee9c4f41c7ff1cea  reserved
557b9fdb9f96569dff7eb2de10d3946f  suspended


test_single_exp-v1
==================
id                                status         min obj
--------------------------------  -----------  ---------
9f360d1b4eb2707f19dd619d0d898dd9  broken
47564e5e390348b9d1335d4013895eb4  completed            0
aefd38473f108016fd4842aa855732ff  interrupted
0695f63ecaf7d78f4b85d4cb344e0dc0  new
b0ea9850c09370215b45b81edd33c7d3  reserved
b49e902aebccce14e834d96e411f896e  suspended


"""

    assert captured == expected


def test_two_related_w_ac(clean_db, family_with_trials, capsys):
    """Test two related experiments with --collapse and --all."""
    orion.core.cli.main(['status', '--collapse', '--all'])

    captured = capsys.readouterr().out

    expected = """\
test_double_exp-v1
==================
id                                status
--------------------------------  -----------
c2187f4954884c801e423d851aec9a0b  broken
e42cc22a15188d72df315b9eac79c9c0  completed
b849f69cc3a77f39382d7435d0d41b14  interrupted
7fbbd152f7ca2c064bf00441e311609d  new
d5f1c1cae188608b581ded20cd198679  new
667513aa2cb2244bee9c4f41c7ff1cea  reserved
557b9fdb9f96569dff7eb2de10d3946f  suspended


"""

    assert captured == expected


def test_three_unrelated_w_ac(clean_db, three_experiments_with_trials, capsys):
    """Test three unrelated experiments with --collapse and --all."""
    orion.core.cli.main(['status', '--collapse', '--all'])

    captured = capsys.readouterr().out

    expected = """\
test_double_exp-v1
==================
id                                status
--------------------------------  -----------
c2187f4954884c801e423d851aec9a0b  broken
e42cc22a15188d72df315b9eac79c9c0  completed
b849f69cc3a77f39382d7435d0d41b14  interrupted
7fbbd152f7ca2c064bf00441e311609d  new
d5f1c1cae188608b581ded20cd198679  new
667513aa2cb2244bee9c4f41c7ff1cea  reserved
557b9fdb9f96569dff7eb2de10d3946f  suspended


test_single_exp-v1
==================
id                                status         min obj
--------------------------------  -----------  ---------
9f360d1b4eb2707f19dd619d0d898dd9  broken
47564e5e390348b9d1335d4013895eb4  completed            0
aefd38473f108016fd4842aa855732ff  interrupted
0695f63ecaf7d78f4b85d4cb344e0dc0  new
b0ea9850c09370215b45b81edd33c7d3  reserved
b49e902aebccce14e834d96e411f896e  suspended


"""

    assert captured == expected


def test_three_related_w_ac(clean_db, three_family_with_trials, capsys):
    """Test three related experiments with --collapse and --all."""
    orion.core.cli.main(['status', '--collapse', '--all'])

    captured = capsys.readouterr().out

    expected = """\
test_double_exp-v1
==================
id                                status
--------------------------------  -----------
c2187f4954884c801e423d851aec9a0b  broken
e42cc22a15188d72df315b9eac79c9c0  completed
b849f69cc3a77f39382d7435d0d41b14  interrupted
7fbbd152f7ca2c064bf00441e311609d  new
d5f1c1cae188608b581ded20cd198679  new
e5bf1dd6dec1a0c690ed62ff9146e5b8  new
667513aa2cb2244bee9c4f41c7ff1cea  reserved
557b9fdb9f96569dff7eb2de10d3946f  suspended


"""

    assert captured == expected


def test_three_related_branch_w_ac(clean_db, three_family_branch_with_trials, capsys):
    """Test three related experiments in a branch with --collapse and --all."""
    orion.core.cli.main(['status', '--collapse', '--all'])

    captured = capsys.readouterr().out

    expected = """\
test_double_exp-v1
==================
id                                status
--------------------------------  -----------
c2187f4954884c801e423d851aec9a0b  broken
e42cc22a15188d72df315b9eac79c9c0  completed
b849f69cc3a77f39382d7435d0d41b14  interrupted
7fbbd152f7ca2c064bf00441e311609d  new
d5f1c1cae188608b581ded20cd198679  new
183148e187a1399989a06ffb02059920  new
667513aa2cb2244bee9c4f41c7ff1cea  reserved
557b9fdb9f96569dff7eb2de10d3946f  suspended


"""

    assert captured == expected


def test_no_experiments_w_name(clean_db, monkeypatch, capsys):
    """Test status when --name <exp> does not exist."""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))
    orion.core.cli.main(['status', '--name', 'test_ghost_exp'])

    captured = capsys.readouterr().out

    assert captured == "No experiment found\n"


def test_experiment_wout_child_w_name(clean_db, unrelated_with_trials, capsys):
    """Test status with the name argument and no child."""
    orion.core.cli.main(['status', '--name', 'test_single_exp'])

    captured = capsys.readouterr().out

    expected = """\
test_single_exp-v1
==================
status         quantity    min obj
-----------  ----------  ---------
broken                1
completed             1          0
interrupted           1
new                   1
reserved              1
suspended             1


"""

    assert captured == expected


def test_experiment_w_child_w_name(clean_db, three_experiments_with_trials, capsys):
    """Test status with the name argument and one child."""
    orion.core.cli.main(['status', '--name', 'test_double_exp'])

    captured = capsys.readouterr().out

    expected = """\
test_double_exp-v1
==================
status         quantity
-----------  ----------
broken                1
completed             1
interrupted           1
new                   1
reserved              1
suspended             1


  test_double_exp_child-v1
  ========================
  status         quantity
  -----------  ----------
  broken                1
  completed             1
  interrupted           1
  new                   1
  reserved              1
  suspended             1


"""

    assert captured == expected


def test_experiment_w_parent_w_name(clean_db, three_experiments_with_trials, capsys):
    """Test status with the name argument and one parent."""
    orion.core.cli.main(['status', '--name', 'test_double_exp_child'])

    captured = capsys.readouterr().out

    expected = """\
test_double_exp_child-v1
========================
status         quantity
-----------  ----------
broken                1
completed             1
interrupted           1
new                   2
reserved              2
suspended             1


"""

    assert captured == expected


def test_experiment_same_name_wout_exv(clean_db, three_experiments_same_name, capsys):
    """Test status with three experiments having the same name but different versions."""
    orion.core.cli.main(['status'])

    captured = capsys.readouterr().out

    expected = """\
test_single_exp-v3
==================
empty


"""

    assert captured == expected


def test_experiment_same_name_wout_exv_w_name(clean_db, three_experiments_same_name, capsys):
    """Test status with three experiments having the same name but different versions."""
    orion.core.cli.main(['status', '--name', 'test_single_exp'])

    captured = capsys.readouterr().out

    expected = """\
test_single_exp-v3
==================
empty


"""

    assert captured == expected


def test_experiment_same_name_wout_exv_w_child_w_name(clean_db,
                                                      three_experiments_family_same_name, capsys):
    """Test status name with two experiments having the same name and one with a child."""
    orion.core.cli.main(['status', '--name', 'test_single_exp'])

    captured = capsys.readouterr().out

    expected = """\
test_single_exp-v1
==================
empty


  test_single_exp-v2
  ==================
  empty


  test_single_exp_child-v1
  ========================
  empty


"""

    assert captured == expected


def test_experiment_same_name_wout_exv_w_c_w_child_w_name(
        clean_db, three_experiments_family_same_name, capsys):
    """Test status name collapsed with two experiments having the same name and one with a child."""
    orion.core.cli.main(['status', '--name', 'test_single_exp', '--collapse'])

    captured = capsys.readouterr().out

    expected = """\
test_single_exp-v2
==================
empty


"""

    assert captured == expected


def test_experiment_same_name_wout_exv_w_child(clean_db,
                                               three_experiments_family_same_name, capsys):
    """Test status with two experiments having the same name and one with a child."""
    orion.core.cli.main(['status'])

    captured = capsys.readouterr().out

    expected = """\
test_single_exp-v1
==================
empty


  test_single_exp-v2
  ==================
  empty


  test_single_exp_child-v1
  ========================
  empty


"""

    assert captured == expected


def test_experiment_same_name_w_exv(clean_db, three_experiments_same_name, capsys):
    """Test status with three experiments with the same name and `--expand-verions`."""
    orion.core.cli.main(['status', '--expand-versions'])

    captured = capsys.readouterr().out

    expected = """\
test_single_exp-v1
==================
empty


  test_single_exp-v2
  ==================
  empty


    test_single_exp-v3
    ==================
    empty


"""

    assert captured == expected


def test_experiment_same_name_w_exv_w_child(clean_db, three_experiments_family_same_name, capsys):
    """Test status with two experiments having the same name and one with a child."""
    orion.core.cli.main(['status', '--expand-versions'])

    captured = capsys.readouterr().out

    expected = """\
test_single_exp-v1
==================
empty


  test_single_exp-v2
  ==================
  empty


  test_single_exp_child-v1
  ========================
  empty


"""

    assert captured == expected


def test_experiment_specific_version(clean_db, three_experiments_same_name, capsys):
    """Test status using `--version`."""
    orion.core.cli.main(['status', '--version', '2'])

    captured = capsys.readouterr().out

    expected = """\
test_single_exp-v2
==================
empty


"""

    assert captured == expected


def test_experiment_cant_use_version(clean_db, three_experiments_same_name):
    """Test status using `--version`."""
    with pytest.raises(RuntimeError) as ex:
        orion.core.cli.main(['status', '--version', '2', '--collapse'])

    assert 'collapse' in str(ex.value)

    with pytest.raises(RuntimeError) as ex:
        orion.core.cli.main(['status', '--version', '2', '--expand-versions'])

    assert 'expand-versions' in str(ex.value)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Collection of tests for :mod:`orion.core.worker.producer`."""
import copy
import datetime
import time

import pytest

from orion.core.io.experiment_builder import build
from orion.core.utils.exceptions import SampleTimeout, WaitingForTrials
from orion.core.utils.format_trials import trial_to_tuple
from orion.core.worker.producer import Producer
from orion.core.worker.trial import Trial


class DumbParallelStrategy:
    """Mock object for parallel strategy"""

    def observe(self, points, results):
        """See BaseParallelStrategy.observe"""
        self._observed_points = points
        self._observed_results = results
        self._value = None

    def lie(self, trial):
        """See BaseParallelStrategy.lie"""
        if self._value:
            value = self._value
        else:
            value = len(self._observed_points)

        self._lie = lie = Trial.Result(name='lie', type='lie', value=value)
        return lie


def produce_lies(producer):
    """Wrap production of lies outside of `Producer.update`"""
    return producer._produce_lies(producer.experiment.fetch_noncompleted_trials())


def update_algorithm(producer):
    """Wrap update of algorithm outside of `Producer.update`"""
    return producer._update_algorithm(producer.experiment.fetch_trials_by_status('completed'))


def update_naive_algorithm(producer):
    """Wrap update of naive algorithm outside of `Producer.update`"""
    return producer._update_naive_algorithm(producer.experiment.fetch_noncompleted_trials())


@pytest.fixture()
def producer(monkeypatch, hacked_exp, random_dt, exp_config, categorical_values):
    """Return a setup `Producer`."""
    # make init done

    hacked_exp.pool_size = 1
    hacked_exp.algorithms.algorithm.possible_values = categorical_values
    hacked_exp.algorithms.seed_rng(0)

    hacked_exp.producer['strategy'] = DumbParallelStrategy()

    producer = Producer(hacked_exp)

    def backoff(self):
        """Dont wait, just update."""
        self.update()
        self.failure_count += 1

    monkeypatch.setattr(Producer, 'backoff', backoff)

    return producer


def test_algo_observe_completed(producer):
    """Test that algo only observes completed trials"""
    assert len(producer.experiment.fetch_trials()) > 3
    producer.update()
    # Algorithm must have received completed points and their results
    obs_points = producer.algorithm.algorithm._points
    obs_results = producer.algorithm.algorithm._results
    assert len(obs_points) == 3
    assert obs_points[0] == ('rnn', 'lstm')
    assert obs_points[1] == ('rnn', 'rnn')
    assert obs_points[2] == ('lstm_with_attention', 'gru')
    assert len(obs_results) == 3
    assert obs_results[0] == {
        'objective': 3,
        'gradient': None,
        'constraint': []
        }
    assert obs_results[1] == {
        'objective': 2,
        'gradient': (-0.1, 2),
        'constraint': []
        }
    assert obs_results[2] == {
        'objective': 10,
        'gradient': (5, 3),
        'constraint': [1.2]
        }


def test_strategist_observe_completed(producer):
    """Test that strategist only observes completed trials"""
    assert len(producer.experiment.fetch_trials()) > 3
    producer.update()
    # Algorithm must have received completed points and their results
    obs_points = producer.strategy._observed_points
    obs_results = producer.strategy._observed_results
    assert len(obs_points) == 3
    assert obs_points[0] == ('rnn', 'lstm')
    assert obs_points[1] == ('rnn', 'rnn')
    assert obs_points[2] == ('lstm_with_attention', 'gru')
    assert len(obs_results) == 3
    assert obs_results[0] == {
        'objective': 3,
        'gradient': None,
        'constraint': []
        }
    assert obs_results[1] == {
        'objective': 2,
        'gradient': (-0.1, 2),
        'constraint': []
        }
    assert obs_results[2] == {
        'objective': 10,
        'gradient': (5, 3),
        'constraint': [1.2]
        }


def test_naive_algorithm_is_producing(monkeypatch, producer, database, random_dt):
    """Verify naive algo is used to produce, not original algo"""
    producer.experiment.pool_size = 1
    producer.algorithm.algorithm.possible_values = [('gru', 'rnn')]
    producer.update()
    monkeypatch.setattr(producer.algorithm.algorithm, 'set_state', lambda value: None)
    producer.algorithm.algorithm.possible_values = [('gru', 'gru')]
    producer.produce()

    assert producer.naive_algorithm.algorithm._num == 1  # pool size
    assert producer.algorithm.algorithm._num == 0


def test_update_and_produce(producer, database, random_dt):
    """Test new trials are properly produced"""
    possible_values = [('gru', 'rnn')]
    producer.experiment.pool_size = 1
    producer.experiment.algorithms.algorithm.possible_values = possible_values

    producer.update()
    producer.produce()

    # Algorithm was ordered to suggest some trials
    num_new_points = producer.naive_algorithm.algorithm._num
    assert num_new_points == 1  # pool size

    assert producer.naive_algorithm.algorithm._suggested == possible_values


def test_register_new_trials(producer, database, random_dt):
    """Test new trials are properly registered"""
    trials_in_db_before = database.trials.count()
    new_trials_in_db_before = database.trials.count({'status': 'new'})

    producer.experiment.pool_size = 1
    producer.experiment.algorithms.algorithm.possible_values = [('gru', 'rnn')]

    producer.update()
    producer.produce()

    # Algorithm was ordered to suggest some trials
    num_new_points = producer.naive_algorithm.algorithm._num
    assert num_new_points == 1  # pool size

    # `num_new_points` new trials were registered at database
    assert database.trials.count() == trials_in_db_before + 1
    assert database.trials.count({'status': 'new'}) == new_trials_in_db_before + 1
    new_trials = list(database.trials.find({'status': 'new', 'submit_time': random_dt}))
    assert new_trials[0]['experiment'] == producer.experiment.name
    assert new_trials[0]['start_time'] is None
    assert new_trials[0]['end_time'] is None
    assert new_trials[0]['results'] == []
    assert new_trials[0]['params'] == [
        {'name': '/decoding_layer',
         'type': 'categorical',
         'value': 'gru'},
        {'name': '/encoding_layer',
         'type': 'categorical',
         'value': 'rnn'}
        ]


def test_no_lies_if_all_trials_completed(producer, database, random_dt):
    """Verify that no lies are created if all trials are completed"""
    query = {'status': {'$ne': 'completed'}, 'experiment': producer.experiment.id}
    database.trials.remove(query)
    trials_in_db_before = database.trials.count({'experiment': producer.experiment.id})
    assert trials_in_db_before == 3

    producer.update()

    assert len(produce_lies(producer)) == 0


def test_lies_generation(producer, database, random_dt):
    """Verify that lies are created properly"""
    query = {'status': {'$ne': 'completed'}, 'experiment': producer.experiment.id}
    trials_non_completed = list(database.trials.find(query))
    assert len(trials_non_completed) == 4
    query = {'status': 'completed', 'experiment': producer.experiment.id}
    trials_completed = list(database.trials.find(query))
    assert len(trials_completed) == 3

    producer.update()

    lies = produce_lies(producer)
    assert len(lies) == 4

    trials_non_completed = list(
        sorted(trials_non_completed,
               key=lambda trial: trial.get('submit_time', datetime.datetime.utcnow())))

    for i in range(4):
        trials_non_completed[i]['_id'] = lies[i].id
        trials_non_completed[i]['status'] = 'completed'
        trials_non_completed[i]['end_time'] = random_dt
        trials_non_completed[i]['results'].append(producer.strategy._lie.to_dict())
        trials_non_completed[i]['parents'] = set([trial['_id'] for trial in trials_completed])
        lies_dict = lies[i].to_dict()
        lies_dict['parents'] = set(lies_dict['parents'])
        assert lies_dict == trials_non_completed[i]


def test_register_lies(producer, database, random_dt):
    """Verify that lies are registed in DB properly"""
    query = {'status': {'$ne': 'completed'}, 'experiment': producer.experiment.id}
    trials_non_completed = list(database.trials.find(query))
    assert len(trials_non_completed) == 4
    query = {'status': 'completed', 'experiment': producer.experiment.id}
    trials_completed = list(database.trials.find(query))
    assert len(trials_completed) == 3

    producer.update()
    produce_lies(producer)

    lying_trials = list(database.lying_trials.find({'experiment': producer.experiment.id}))
    assert len(lying_trials) == 4

    trials_non_completed = list(
        sorted(trials_non_completed,
               key=lambda trial: trial.get('submit_time', datetime.datetime.utcnow())))

    for i in range(4):
        trials_non_completed[i]['_id'] = lying_trials[i]['_id']
        trials_non_completed[i]['status'] = 'completed'
        trials_non_completed[i]['end_time'] = random_dt
        trials_non_completed[i]['results'].append(producer.strategy._lie.to_dict())
        trials_non_completed[i]['parents'] = set([trial['_id'] for trial in trials_completed])
        lying_trials[i]['parents'] = set(lying_trials[i]['parents'])
        assert lying_trials[i] == trials_non_completed[i]


def test_register_duplicate_lies(producer, database, random_dt):
    """Verify that duplicate lies are not registered twice in DB"""
    query = {'status': {'$ne': 'completed'}, 'experiment': producer.experiment.id}
    trials_non_completed = list(database.trials.find(query))
    assert len(trials_non_completed) == 4

    # Overwrite value of lying result of the strategist so that all lying trials have the same value
    # otherwise they would not be exact duplicates.
    producer.strategy._value = 4

    # Set specific output value for to algo to ensure successful creation of a new trial.
    producer.experiment.pool_size = 1
    producer.experiment.algorithms.algorithm.possible_values = [('gru', 'rnn')]

    producer.update()
    assert len(produce_lies(producer)) == 4
    lying_trials = list(database.lying_trials.find({'experiment': producer.experiment.id}))
    assert len(lying_trials) == 4

    # Create a new point to make sure additional non-completed trials increase number of lying
    # trials generated
    producer.produce()

    trials_non_completed = list(database.trials.find(query))
    assert len(trials_non_completed) == 5

    producer.update()

    assert len(produce_lies(producer)) == 5
    lying_trials = list(database.lying_trials.find({'experiment': producer.experiment.id}))
    assert len(lying_trials) == 5

    # Make sure trying to generate again does not add more fake trials since they are identical
    assert len(produce_lies(producer)) == 5
    lying_trials = list(database.lying_trials.find({'experiment': producer.experiment.id}))
    assert len(lying_trials) == 5


def test_register_duplicate_lies_with_different_results(producer, database, random_dt):
    """Verify that duplicate lies with different results are all registered in DB"""
    query = {'status': {'$ne': 'completed'}, 'experiment': producer.experiment.id}
    trials_non_completed = list(database.trials.find(query))
    assert len(trials_non_completed) == 4

    # Overwrite value of lying result to force different results.
    producer.strategy._value = 11

    assert len(produce_lies(producer)) == 4
    lying_trials = list(database.lying_trials.find({'experiment': producer.experiment.id}))
    assert len(lying_trials) == 4

    # Overwrite value of lying result to force different results.
    producer.strategy._value = new_lying_value = 12

    lying_trials = produce_lies(producer)
    assert len(lying_trials) == 4
    nb_lying_trials = database.lying_trials.count({'experiment': producer.experiment.id})
    assert nb_lying_trials == 4 + 4
    assert lying_trials[0].lie.value == new_lying_value


def test_naive_algo_not_trained_when_all_trials_completed(producer, database, random_dt):
    """Verify that naive algo is not trained on additional trials when all completed"""
    query = {'status': {'$ne': 'completed'}, 'experiment': producer.experiment.id}
    database.trials.remove(query)
    trials_in_db_before = database.trials.count({'experiment': producer.experiment.id})
    assert trials_in_db_before == 3

    producer.update()

    assert len(producer.algorithm.algorithm._points) == 3
    assert len(producer.naive_algorithm.algorithm._points) == 3


def test_naive_algo_trained_on_all_non_completed_trials(producer, database, random_dt):
    """Verify that naive algo is trained on additional trials"""
    # Set two of completed trials to broken and reserved to have all possible status
    query = {'status': 'completed', 'experiment': producer.experiment.id}
    completed_trials = database.trials.find(query)
    database.trials.update({'_id': completed_trials[0]['_id']}, {'$set': {'status': 'broken'}})
    database.trials.update({'_id': completed_trials[1]['_id']}, {'$set': {'status': 'reserved'}})

    # Make sure non completed trials and completed trials are set properly for the unit-test
    query = {'status': {'$ne': 'completed'}, 'experiment': producer.experiment.id}
    non_completed_trials = list(database.trials.find(query))
    assert len(non_completed_trials) == 6
    # Make sure we have all type of status except completed
    assert (set(trial['status'] for trial in non_completed_trials) ==
            set(['new', 'reserved', 'suspended', 'interrupted', 'broken']))
    query = {'status': 'completed', 'experiment': producer.experiment.id}
    assert database.trials.count(query) == 1

    # Executing the actual test
    producer.update()
    assert len(produce_lies(producer)) == 6

    assert len(producer.algorithm.algorithm._points) == 1
    assert len(producer.naive_algorithm.algorithm._points) == (1 + 6)


def test_naive_algo_is_discared(producer, database, monkeypatch):
    """Verify that naive algo is discarded and recopied from original algo"""
    # Set values for predictions
    producer.experiment.pool_size = 1
    producer.experiment.algorithms.algorithm.possible_values = [('gru', 'rnn')]

    producer.update()
    assert len(produce_lies(producer)) == 4

    first_naive_algorithm = producer.naive_algorithm

    assert len(producer.algorithm.algorithm._points) == 3
    assert len(first_naive_algorithm.algorithm._points) == (3 + 4)

    producer.produce()

    # Only update the original algo, naive algo is still not discarded
    update_algorithm(producer)
    assert len(producer.algorithm.algorithm._points) == 3
    assert first_naive_algorithm == producer.naive_algorithm
    assert len(producer.naive_algorithm.algorithm._points) == (3 + 4)

    # Discard naive algo and create a new one, now trained on 5 points.
    update_naive_algorithm(producer)
    assert first_naive_algorithm != producer.naive_algorithm
    assert len(producer.naive_algorithm.algorithm._points) == (3 + 5)


def test_concurent_producers(producer, database, random_dt):
    """Test concurrent production of new trials."""
    trials_in_db_before = database.trials.count()
    new_trials_in_db_before = database.trials.count({'status': 'new'})

    # Set so that first producer's algorithm generate valid point on first time
    # And second producer produce same point and thus must produce next one two.
    # Hence, we know that producer algo will have _num == 1 and
    # second producer algo will have _num == 2
    producer.algorithm.algorithm.possible_values = [('gru', 'rnn'), ('gru', 'gru')]
    # Make sure it starts from index 0
    producer.algorithm.seed_rng(0)

    assert producer.experiment.pool_size == 1

    second_producer = Producer(producer.experiment)
    second_producer.algorithm = copy.deepcopy(producer.algorithm)

    producer.update()
    second_producer.update()

    producer.produce()
    second_producer.produce()

    # Algorithm was required to suggest some trials
    num_new_points = producer.algorithm.algorithm._num
    assert num_new_points == 1  # pool size
    num_new_points = second_producer.algorithm.algorithm._num
    assert num_new_points == 2  # pool size

    # `num_new_points` new trials were registered at database
    assert database.trials.count() == trials_in_db_before + 2
    assert database.trials.count({'status': 'new'}) == new_trials_in_db_before + 2
    new_trials = list(database.trials.find({'status': 'new', 'submit_time': random_dt}))
    assert new_trials[0]['experiment'] == producer.experiment.name
    assert new_trials[0]['start_time'] is None
    assert new_trials[0]['end_time'] is None
    assert new_trials[0]['results'] == []
    assert new_trials[0]['params'] == [
        {'name': '/decoding_layer',
         'type': 'categorical',
         'value': 'gru'},
        {'name': '/encoding_layer',
         'type': 'categorical',
         'value': 'rnn'}
        ]

    assert new_trials[1]['params'] == [
        {'name': '/decoding_layer',
         'type': 'categorical',
         'value': 'gru'},
        {'name': '/encoding_layer',
         'type': 'categorical',
         'value': 'gru'}
        ]


def test_duplicate_within_pool(producer, database, random_dt):
    """Test that an algo suggesting multiple points can have a few registered even
    if one of them is a duplicate.
    """
    trials_in_db_before = database.trials.count()
    new_trials_in_db_before = database.trials.count({'status': 'new'})

    producer.experiment.pool_size = 2

    producer.experiment.algorithms.algorithm.possible_values = [
        ('gru', 'rnn'), ('gru', 'rnn'), ('gru', 'gru')]

    producer.update()
    producer.produce()

    # Algorithm was required to suggest some trials
    num_new_points = producer.algorithm.algorithm._num
    assert num_new_points == 4  # 2 * pool size

    # `num_new_points` new trials were registered at database
    assert database.trials.count() == trials_in_db_before + 2
    assert database.trials.count({'status': 'new'}) == new_trials_in_db_before + 2
    new_trials = list(database.trials.find({'status': 'new', 'submit_time': random_dt}))
    assert new_trials[0]['experiment'] == producer.experiment.name
    assert new_trials[0]['start_time'] is None
    assert new_trials[0]['end_time'] is None
    assert new_trials[0]['results'] == []
    assert new_trials[0]['params'] == [
        {'name': '/decoding_layer',
         'type': 'categorical',
         'value': 'gru'},
        {'name': '/encoding_layer',
         'type': 'categorical',
         'value': 'rnn'}
        ]

    assert new_trials[1]['params'] == [
        {'name': '/decoding_layer',
         'type': 'categorical',
         'value': 'gru'},
        {'name': '/encoding_layer',
         'type': 'categorical',
         'value': 'gru'}
        ]


def test_duplicate_within_pool_and_db(producer, database, random_dt):
    """Test that an algo suggesting multiple points can have a few registered even
    if one of them is a duplicate with db.
    """
    trials_in_db_before = database.trials.count()
    new_trials_in_db_before = database.trials.count({'status': 'new'})

    producer.experiment.pool_size = 2

    producer.experiment.algorithms.algorithm.possible_values = [
        ('gru', 'rnn'), ('rnn', 'rnn'), ('gru', 'gru')]

    producer.update()
    producer.produce()

    # Algorithm was required to suggest some trials
    num_new_points = producer.algorithm.algorithm._num
    assert num_new_points == 4  # pool size

    # `num_new_points` new trials were registered at database
    assert database.trials.count() == trials_in_db_before + 2
    assert database.trials.count({'status': 'new'}) == new_trials_in_db_before + 2
    new_trials = list(database.trials.find({'status': 'new', 'submit_time': random_dt}))
    assert new_trials[0]['experiment'] == producer.experiment.name
    assert new_trials[0]['start_time'] is None
    assert new_trials[0]['end_time'] is None
    assert new_trials[0]['results'] == []
    assert new_trials[0]['params'] == [
        {'name': '/decoding_layer',
         'type': 'categorical',
         'value': 'gru'},
        {'name': '/encoding_layer',
         'type': 'categorical',
         'value': 'rnn'}
        ]

    assert new_trials[1]['params'] == [
        {'name': '/decoding_layer',
         'type': 'categorical',
         'value': 'gru'},
        {'name': '/encoding_layer',
         'type': 'categorical',
         'value': 'gru'}
        ]


def test_exceed_max_idle_time_because_of_duplicates(producer, database, random_dt):
    """Test that RuntimeError is raised when algo keep suggesting the same points"""
    timeout = 3
    producer.max_idle_time = timeout  # to limit run-time, default would work as well.
    producer.experiment.algorithms.algorithm.possible_values = [('rnn', 'rnn')]

    assert producer.experiment.pool_size == 1

    producer.update()

    start = time.time()

    with pytest.raises(SampleTimeout):
        producer.produce()

    assert timeout <= time.time() - start < timeout + 1


def test_exceed_max_idle_time_because_of_optout(producer, database, random_dt, monkeypatch):
    """Test that RuntimeError is raised when algo keeps opting out"""
    timeout = 3
    producer.max_idle_time = timeout  # to limit run-time, default would work as well.

    def opt_out(self, num=1):
        """Return None to always opt out"""
        self._suggested = None
        return None

    monkeypatch.setattr(producer.experiment.algorithms.algorithm.__class__, 'suggest', opt_out)

    assert producer.experiment.pool_size == 1

    producer.update()

    with pytest.raises(WaitingForTrials):
        producer.produce()


def test_stops_if_algo_done(producer, database, random_dt, monkeypatch):
    """Test that producer stops producing when algo is done."""
    def opt_out_and_complete(self, num=1):
        """Return None to always opt out and set id_done to True"""
        self._suggested = None
        self.done = True
        print('set')
        return None

    monkeypatch.setattr(producer.experiment.algorithms.algorithm.__class__, 'suggest',
                        opt_out_and_complete)

    assert producer.experiment.pool_size == 1

    trials_in_db_before = database.trials.count()

    producer.update()
    producer.produce()

    assert database.trials.count() == trials_in_db_before
    assert producer.experiment.algorithms.is_done


def test_original_seeding(producer, database):
    """Verify that rng state in original algo changes when duplicate trials is discarded"""
    assert producer.experiment.pool_size == 1

    producer.algorithm.seed_rng(0)

    assert producer.algorithm.algorithm._index == 0

    producer.update()
    producer.produce()

    prev_index = producer.algorithm.algorithm._index
    prev_suggested = producer.algorithm.algorithm._suggested
    assert prev_index > 0

    # Force the algo back to 1 to make sure the RNG state of original algo keeps incrementing.
    # This is necessary because naive_algo is recopied from original algo and thus would always get
    # the same RNG state if the original algo RNG state would not increment.
    # See `Producer.produce` to observe the dummy `self.algorith.suggest()` used to increment
    # original algo's RNG state.
    producer.algorithm.seed_rng(0)

    producer.update()
    producer.produce()

    assert prev_suggested != producer.algorithm.algorithm._suggested
    assert prev_index < producer.algorithm.algorithm._index


def test_evc(monkeypatch, producer):
    """Verify that producer is using available trials from EVC"""
    experiment = producer.experiment
    new_experiment = build(experiment.name, algorithms='random')

    # Replace parent with hacked exp, otherwise parent ID does not match trials in DB
    # and fetch_trials() won't return anything.
    new_experiment._node.parent._item = experiment

    assert len(new_experiment.fetch_trials(with_evc_tree=True)) == len(experiment.fetch_trials())

    producer.experiment = new_experiment

    def update_algo(trials):
        assert len(trials) == 3

    def update_naive_algo(trials):
        assert len(trials) == 4

    monkeypatch.setattr(producer, '_update_algorithm', update_algo)
    monkeypatch.setattr(producer, '_update_naive_algorithm', update_naive_algo)

    producer.update()


def test_evc_duplicates(monkeypatch, producer):
    """Verify that producer wont register samples that are available in parent experiment"""
    experiment = producer.experiment
    new_experiment = build(experiment.name, algorithms='random')

    # Replace parent with hacked exp, otherwise parent ID does not match trials in DB
    # and fetch_trials() won't return anything.
    new_experiment._node.parent._item = experiment

    assert len(new_experiment.fetch_trials(with_evc_tree=True)) == len(experiment.fetch_trials())

    def suggest(pool_size):
        return [trial_to_tuple(experiment.fetch_trials()[-1], experiment.space)]

    producer.experiment = new_experiment
    producer.algorithm = new_experiment.algorithms
    producer.max_idle_time = 1

    monkeypatch.setattr(new_experiment.algorithms, 'suggest', suggest)

    producer.update()
    with pytest.raises(SampleTimeout):
        producer.produce()

    assert len(new_experiment.fetch_trials(with_evc_tree=False)) == 0


def test_algorithm_is_done(monkeypatch, producer):
    """Verify that producer won't register new samples if algorithm is done meanwhile."""
    producer.experiment.max_trials = 8
    producer.experiment.pool_size = 10
    # Reset Producer to test that max_trial is set properly during init.
    producer = Producer(producer.experiment)

    def suggest_one_only(self, num=1):
        """Return only one point, whatever `num` is"""
        return [('gru', 'rnn')]

    monkeypatch.delattr(producer.experiment.algorithms.algorithm.__class__, 'is_done')
    monkeypatch.setattr(producer.experiment.algorithms.algorithm.__class__, 'suggest',
                        suggest_one_only)

    assert producer.experiment.pool_size == 10
    trials_in_exp_before = len(producer.experiment.fetch_trials())
    assert trials_in_exp_before == producer.experiment.max_trials - 1

    producer.update()
    producer.produce()

    assert len(producer.experiment.fetch_trials()) == producer.experiment.max_trials
    assert producer.naive_algorithm.is_done
    assert not producer.experiment.is_done



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Example usage and tests for :mod:`orion.core.utils.format`."""

import pytest

from orion.core.utils.format_trials import (dict_to_trial, trial_to_tuple, tuple_to_trial)
from orion.core.worker.trial import Trial


@pytest.fixture()
def trial():
    """Stab trial to match tuple from fixture `fixed_suggestion`."""
    params = [
        dict(
            name='yolo',
            type='categorical',
            value=('asdfa', 2)
            ),
        dict(
            name='yolo2',
            type='integer',
            value=0
            ),
        dict(
            name='yolo3',
            type='real',
            value=3.5
            )
        ]
    return Trial(params=params)


@pytest.fixture()
def hierarchical_trial():
    """Stab trial with hierarchical params."""
    params = [
        dict(
            name='yolo.first',
            type='categorical',
            value=('asdfa', 2)
            ),
        dict(
            name='yolo.second',
            type='integer',
            value=0
            ),
        dict(
            name='yoloflat',
            type='real',
            value=3.5
            )
        ]
    return Trial(params=params)


@pytest.fixture()
def dict_params():
    """Return dictionary of params to build a trial like `fixed_suggestion`"""
    return {'yolo': ('asdfa', 2), 'yolo2': 0, 'yolo3': 3.5}


@pytest.fixture()
def hierarchical_dict_params():
    """Return dictionary of params to build a hierarchical trial"""
    return {'yolo': {'first': ('asdfa', 2), 'second': 0}, 'yoloflat': 3.5}


def test_trial_to_tuple(space, trial, fixed_suggestion):
    """Check if trial is correctly created from a sample/tuple."""
    data = trial_to_tuple(trial, space)
    assert data == fixed_suggestion

    trial._params[0].name = 'lalala'
    with pytest.raises(ValueError) as exc:
        trial_to_tuple(trial, space)

    assert "Trial params: [\'lalala\', \'yolo2\', \'yolo3\']" in str(exc.value)

    trial._params.pop(0)
    with pytest.raises(ValueError) as exc:
        trial_to_tuple(trial, space)

    assert "Trial params: [\'yolo2\', \'yolo3\']" in str(exc.value)


def test_tuple_to_trial(space, trial, fixed_suggestion):
    """Check if sample is recovered successfully from trial."""
    t = tuple_to_trial(fixed_suggestion, space)
    assert t.experiment is None
    assert t.status == 'new'
    assert t.worker is None
    assert t.submit_time is None
    assert t.start_time is None
    assert t.end_time is None
    assert t.results == []
    assert len(t._params) == len(trial.params)
    for i in range(len(t.params)):
        assert t._params[i].to_dict() == trial._params[i].to_dict()


def test_dict_to_trial(space, trial, dict_params):
    """Check if dict is converted successfully to trial."""
    t = dict_to_trial(dict_params, space)
    assert t.experiment is None
    assert t.status == 'new'
    assert t.worker is None
    assert t.submit_time is None
    assert t.start_time is None
    assert t.end_time is None
    assert t.results == []
    assert len(t._params) == len(trial._params)
    for i in range(len(t.params)):
        assert t._params[i].to_dict() == trial._params[i].to_dict()


def test_tuple_to_trial_to_tuple(space, trial, fixed_suggestion):
    """The two functions should be inverse."""
    data = trial_to_tuple(tuple_to_trial(fixed_suggestion, space), space)
    assert data == fixed_suggestion

    t = tuple_to_trial(trial_to_tuple(trial, space), space)
    assert t.experiment is None
    assert t.status == 'new'
    assert t.worker is None
    assert t.submit_time is None
    assert t.start_time is None
    assert t.end_time is None
    assert t.results == []
    assert len(t._params) == len(trial._params)
    for i in range(len(t._params)):
        assert t._params[i].to_dict() == trial._params[i].to_dict()


def test_hierarchical_trial_to_tuple(hierarchical_space, hierarchical_trial, fixed_suggestion):
    """Check if hierarchical trial is correctly created from a sample/tuple."""
    data = trial_to_tuple(hierarchical_trial, hierarchical_space)
    assert data == fixed_suggestion


def test_tuple_to_hierarchical_trial(hierarchical_space, hierarchical_trial, fixed_suggestion):
    """Check if sample is recovered successfully from hierarchical trial."""
    t = tuple_to_trial(fixed_suggestion, hierarchical_space)
    assert len(t._params) == len(hierarchical_trial._params)
    for i in range(len(t._params)):
        assert t._params[i].to_dict() == hierarchical_trial._params[i].to_dict()


def test_hierarchical_dict_to_trial(hierarchical_space, hierarchical_trial,
                                    hierarchical_dict_params):
    """Check if hierarchical dict is converted successfully to trial."""
    t = dict_to_trial(hierarchical_dict_params, hierarchical_space)
    assert len(t._params) == len(hierarchical_trial._params)
    for i in range(len(t.params)):
        assert t._params[i].to_dict() == hierarchical_trial._params[i].to_dict()



# -*- coding: utf-8 -*-
"""
:mod:`orion.algo.evolution_es` --
The Evolved Transformer and large-scale evolution of image classifiers
===========================================================================================

.. module:: evolution_es
    :platform: Unix
    :synopsis: Implement evolution to exploit configurations with fixed resource efficiently

"""
import copy
import importlib
import logging

import numpy as np

from orion.algo.hyperband import Bracket, Hyperband

logger = logging.getLogger(__name__)

REGISTRATION_ERROR = """
Bad fidelity level {fidelity}. Should be in {budgets}.
Params: {params}
"""

SPACE_ERROR = """
EvolutionES cannot be used if space does not contain a fidelity dimension.
"""

BUDGET_ERROR = """
Cannot build budgets below max_resources;
(max: {}) - (min: {}) > (num_rungs: {})
"""


def compute_budgets(min_resources, max_resources, reduction_factor, nums_population, pairs):
    """Compute the budgets used for each execution of hyperband"""
    budgets_eves = []
    if reduction_factor == 1:
        for i in range(min_resources, max_resources + 1):
            if i == min_resources:
                budgets_eves.append([(nums_population, i)])
            else:
                budgets_eves[0].append((pairs * 2, i))
    else:
        num_brackets = int(np.log(max_resources) / np.log(reduction_factor))
        budgets = []
        budgets_tab = {}  # just for display consideration
        for bracket_id in range(0, num_brackets + 1):
            bracket_budgets = []
            num_trials = int(np.ceil(int((num_brackets + 1) / (num_brackets - bracket_id + 1)) *
                                     (reduction_factor ** (num_brackets - bracket_id))))

            min_resources = max_resources / reduction_factor ** (num_brackets - bracket_id)
            for i in range(0, num_brackets - bracket_id + 1):
                n_i = int(num_trials / reduction_factor ** i)
                min_i = int(min_resources * reduction_factor ** i)
                bracket_budgets.append((n_i, min_i))

                if budgets_tab.get(i):
                    budgets_tab[i].append((n_i, min_i))
                else:
                    budgets_tab[i] = [(n_i, min_i)]

            budgets.append(bracket_budgets)

        for i in range(len(budgets[0])):
            if i == 0:
                budgets_eves.append([(nums_population, budgets[0][i][1])])
            else:
                budgets_eves[0].append((pairs * 2, budgets[0][i][1]))

    return budgets_eves


class EvolutionES(Hyperband):
    """EvolutionES formulates hyperparameter optimization as an evolution.

    For more information on the algorithm,
    see original paper at
    https://arxiv.org/pdf/1703.01041.pdf and
    https://arxiv.org/pdf/1901.11117.pdf

    Real et al. "Large-Scale Evolution of Image Classifiers"
    So et all. "The Evolved Transformer"

    Parameters
    ----------
    space: `orion.algo.space.Space`
        Optimisation space with priors for each dimension.
    seed: None, int or sequence of int
        Seed for the random number generator used to sample new trials.
        Default: ``None``
    repetitions: int
        Number of execution of Hyperband. Default is numpy.inf which means to
        run Hyperband until no new trials can be suggested.

    """

    def __init__(self, space, seed=None, repetitions=np.inf, nums_population=20, mutate=None):
        super(EvolutionES, self).__init__(space, seed=seed, repetitions=repetitions)
        pair = nums_population // 2
        mutate_ratio = 0.3
        self.nums_population = nums_population
        self.nums_comp_pairs = pair
        self.mutate_ratio = mutate_ratio
        self.mutate_attr = mutate
        self.nums_mutate_gene = int((len(self.space.values()) - 1) * mutate_ratio) if int(
            (len(self.space.values()) - 1) * mutate_ratio) > 0 else 1

        self.hurdles = []

        self.population = {}
        for key in range(len(self.space)):
            if not key == self.fidelity_index:
                self.population[key] = -1 * np.ones(nums_population)

        self.performance = np.inf * np.ones(nums_population)

        self.budgets = compute_budgets(self.min_resources, self.max_resources,
                                       self.reduction_factor, nums_population, pair)

        self.brackets = [
            BracketEVES(self, bracket_budgets, 1, space)
            for bracket_budgets in self.budgets
        ]
        self.seed_rng(seed)

    def _get_bracket(self, point):
        """Get the bracket of a point during observe"""
        return self.brackets[0]


class BracketEVES(Bracket):
    """Bracket of rungs for the algorithm Hyperband.

    Parameters
    ----------
    evolutiones: `evolutiones` algorithm
        The evolutiones algorithm object which this bracket will be part of.
    budgets: list of tuple
        Each tuple gives the (n_trials, resource_budget) for the respective rung.
    repetition_id: int
        The id of hyperband execution this bracket belongs to

    """

    def __init__(self, evolution_es, budgets, repetition_id, space):
        super(BracketEVES, self).__init__(evolution_es, budgets, repetition_id)
        self.eves = self.hyperband
        self.space = space
        self.search_space_remove_fidelity = []

        if evolution_es.mutate_attr:
            self.mutate_attr = copy.deepcopy(evolution_es.mutate_attr)
        else:
            self.mutate_attr = {}

        function_string = self.mutate_attr.pop('function',
                                               "orion.algo.mutate_functions.default_mutate")
        mod_name, func_name = function_string.rsplit('.', 1)
        mod = importlib.import_module(mod_name)
        self.mutate_func = getattr(mod, func_name)

        for i in range(len(space.values())):
            if not i == self.eves.fidelity_index:
                self.search_space_remove_fidelity.append(i)

    def _get_teams(self, rung_id):
        """Get the red team and blue team"""
        if self.has_rung_filled(rung_id + 1):
            return []

        rung = self.rungs[rung_id]['results']

        population_range = (self.eves.nums_population
                            if len(list(rung.values())) > self.eves.nums_population
                            else len(list(rung.values())))

        for i in range(population_range):
            for j in self.search_space_remove_fidelity:
                self.eves.population[j][i] = list(rung.values())[i][1][j]
            self.eves.performance[i] = list(rung.values())[i][0]

        population_index = list(range(self.eves.nums_population))
        red_team = self.eves.rng.choice(population_index, self.eves.nums_comp_pairs,
                                        replace=False)
        diff_list = list(set(population_index).difference(set(red_team)))
        blue_team = self.eves.rng.choice(diff_list, self.eves.nums_comp_pairs,
                                         replace=False)

        return rung, population_range, red_team, blue_team

    def _mutate_population(self, red_team, blue_team, rung, population_range):
        """Get the mutated population and hurdles"""
        winner_list = []
        loser_list = []

        if set(red_team) != set(blue_team):
            hurdles = 0
            for i, _ in enumerate(red_team):
                winner, loser = ((red_team, blue_team)
                                 if self.eves.performance[red_team[i]] <
                                 self.eves.performance[blue_team[i]]
                                 else (blue_team, red_team))

                winner_list.append(winner[i])
                loser_list.append(loser[i])
                hurdles += self.eves.performance[winner[i]]
                self._mutate(winner[i], loser[i])

            hurdles /= len(red_team)
            self.eves.hurdles.append(hurdles)

            logger.debug('Evolution hurdles are: %s', str(self.eves.hurdles))

        points = []
        nums_all_equal = [0] * population_range
        for i in range(population_range):
            point = [0] * len(self.space)
            while True:
                point[self.eves.fidelity_index] = \
                    list(rung.values())[i][1][self.eves.fidelity_index]

                for j in self.search_space_remove_fidelity:
                    point[j] = self.eves.population[j][i]

                if tuple(point) in points:
                    nums_all_equal[i] += 1
                    logger.debug("find equal one, continue to mutate.")
                    self._mutate(points.index(tuple(point)), i)
                else:
                    break
                if nums_all_equal[i] > 10:
                    logger.warning("Can not Evolve any more, "
                                   "You can make an early stop.")
                    break

            points.append(tuple(point))

        return points, np.array(nums_all_equal)

    def get_candidates(self, rung_id):
        """Get a candidate for promotion"""
        rung, population_range, red_team, blue_team = self._get_teams(rung_id)
        return self._mutate_population(red_team, blue_team, rung, population_range)[0]

    def _mutate(self, winner_id, loser_id):
        select_genes_key_list = self.eves.rng.choice(self.search_space_remove_fidelity,
                                                     self.eves.nums_mutate_gene,
                                                     replace=False)
        self.copy_winner(winner_id, loser_id)
        kwargs = copy.deepcopy(self.mutate_attr)

        for i, _ in enumerate(select_genes_key_list):
            space = self.space.values()[select_genes_key_list[i]]
            old = self.eves.population[select_genes_key_list[i]][loser_id]
            new = self.mutate_func(space, old, **kwargs)
            self.eves.population[select_genes_key_list[i]][loser_id] = new

        self.eves.performance[loser_id] = -1

    def copy_winner(self, winner_id, loser_id):
        """Copy winner to loser"""
        for key in self.search_space_remove_fidelity:
            self.eves.population[key][loser_id] = self.eves.population[key][winner_id].copy()



# -*- coding: utf-8 -*-
"""
:mod:`orion.testing` -- Common testing support module
=====================================================
.. module:: testing
   :platform: Unix
   :synopsis: Common testing support module providing defaults, functions and mocks.
"""
# pylint: disable=protected-access

from contextlib import contextmanager
import copy
import datetime

import orion.core.io.experiment_builder as experiment_builder
from orion.core.worker.producer import Producer
from orion.testing.state import OrionState


def default_datetime():
    """Return default datetime"""
    return datetime.datetime(1903, 4, 25, 0, 0, 0)


def generate_trials(trial_config, statuses):
    """Generate Trials with different configurations"""

    def _generate(obj, *args, value):
        if obj is None:
            return None

        obj = copy.deepcopy(obj)
        data = obj

        data[args[-1]] = value
        return obj

    new_trials = [_generate(trial_config, 'status', value=s) for s in statuses]

    for i, trial in enumerate(new_trials):
        trial['submit_time'] = datetime.datetime.utcnow() + datetime.timedelta(seconds=i)
        if trial['status'] != 'new':
            trial['start_time'] = datetime.datetime.utcnow() + datetime.timedelta(seconds=i)

    for i, trial in enumerate(new_trials):
        if trial['status'] == 'completed':
            trial['end_time'] = datetime.datetime.utcnow() + datetime.timedelta(seconds=i)

    # make each trial unique
    for i, trial in enumerate(new_trials):
        if trial['status'] == 'completed':
            trial['results'].append({
                'name': 'loss',
                'type': 'objective',
                'value': i})

        trial['params'].append({
            'name': 'x',
            'type': 'real',
            'value': i
        })

    return new_trials


@contextmanager
def create_experiment(exp_config=None, trial_config=None, statuses=None):
    """Context manager for the creation of an ExperimentClient and storage init"""
    if exp_config is None:
        raise ValueError("Parameter 'exp_config' is missing")
    if trial_config is None:
        raise ValueError("Parameter 'trial_config' is missing")
    if statuses is None:
        statuses = ['new', 'interrupted', 'suspended', 'reserved', 'completed']

    from orion.client.experiment import ExperimentClient

    with OrionState(experiments=[exp_config],
                    trials=generate_trials(trial_config, statuses)) as cfg:
        experiment = experiment_builder.build(name=exp_config['name'])
        if cfg.trials:
            experiment._id = cfg.trials[0]['experiment']
        client = ExperimentClient(experiment, Producer(experiment))
        yield cfg, experiment, client

    client.close()


class MockDatetime(datetime.datetime):
    """Fake Datetime"""

    @classmethod
    def utcnow(cls):
        """Return our random/fixed datetime"""
        return default_datetime()



# -*- coding: utf-8 -*-
"""
:mod:`orion.core.utils.working_dir` -- ContextManager for working directory
===========================================================================

.. module:: working_dir
   :platform: Unix
   :synopsis: ContextManager class to create a permanent directory or a temporary one.
"""
import os
import tempfile


# pylint: disable=too-few-public-methods
class WorkingDir:
    """ContextManager class for temporary or permanent directory."""

    def __init__(self, working_dir, temp=True, suffix=None, prefix=None):
        """Create the context manager with the given name.

        Parameters
        ----------
        name : str, optional
            Name of the directory. If empty, will create a temporary one.

        """
        self.working_dir = str(working_dir)
        self._temp = temp
        self._suffix = suffix
        self._prefix = prefix
        self._tmpdir = None

    def __enter__(self):
        """Create the a permanent directory or a temporary one."""
        os.makedirs(self.working_dir, exist_ok=True)

        if not self._temp:
            path = os.path.join(self.working_dir, self._prefix + self._suffix)
            os.makedirs(path, exist_ok=True)
            return path

        self._tmpdir = tempfile.TemporaryDirectory(suffix=self._suffix, prefix=self._prefix,
                                                   dir=self.working_dir)
        return self._tmpdir.name

    def __exit__(self, exc_type, exc_value, traceback):
        """Cleanup temporary directory."""
        if self._temp:
            self._tmpdir.cleanup()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
:mod:`orion.core.cli.db_main` -- Module containing database related operations
==============================================================================

.. module:: db_main
   :platform: Unix
   :synopsis: Root command for database operations

"""
import logging

from orion.core.utils import module_import

log = logging.getLogger(__name__)
SHORT_DESCRIPTION = 'Database initialization, upgrade, verification, and edition'
DESCRIPTION = """
Root command for database operations.
"""


def add_subparser(parser):
    """Add the subparsers that needs to be used for this command"""
    # Fetch experiment name, user's script path and command line arguments
    # Use `-h` option to show help

    db_parser = parser.add_parser('db', help=SHORT_DESCRIPTION, description=DESCRIPTION)
    subparsers = db_parser.add_subparsers()

    load_modules_parser(subparsers)

    return db_parser


def load_modules_parser(subparsers):
    """Search through the `cli.db` folder for any module containing a `get_parser` function"""
    modules = module_import.load_modules_in_path('orion.core.cli.db',
                                                 lambda m: hasattr(m, 'add_subparser'))

    for module in modules:
        get_parser = getattr(module, 'add_subparser')
        get_parser(subparsers)



# -*- coding: utf-8 -*-
"""
:mod:`orion.core.utils.format_trials` -- Utility functions for formatting data
==============================================================================

.. module:: format_trials
   :platform: Unix
   :synopsis: Conversion functions between various data types used in
      framework's ecosystem.

"""

from orion.core.utils.flatten import flatten
from orion.core.worker.trial import Trial


def trial_to_tuple(trial, space):
    """Extract a parameter tuple from a `orion.core.worker.trial.Trial`.

    The order within the tuple is dictated by the defined
    `orion.algo.space.Space` object.
    """
    params = flatten(trial.params)
    trial_keys = set(params.keys())
    space_keys = set(space.keys())
    if trial_keys != space_keys:
        raise ValueError(""""
The trial {} has wrong params:
Trial params: {}
Space dims: {}""".format(trial.id, sorted(trial_keys), sorted(space_keys)))
    return tuple(params[name] for name in space.keys())


def dict_to_trial(data, space):
    """Create a `orion.core.worker.trial.Trial` object from `data`,
    filling only parameter information from `data`.

    :param data: A dict representing a sample point from `space`.
    :param space: Definition of problem's domain.
    :type space: `orion.algo.space.Space`
    """
    data = flatten(data)
    params = []
    for name, dim in space.items():
        if name not in data and dim.default_value is dim.NO_DEFAULT_VALUE:
            raise ValueError(
                'Dimension {} not specified and does not have a default value.'.format(name))
        value = data.get(name, dim.default_value)

        if value not in dim:
            error_msg = "Dimension {} value {} is outside of prior {}".format(
                name, value, dim.get_prior_string())
            raise ValueError(error_msg)

        params.append(dict(
            name=dim.name,
            type=dim.type,
            value=value
            ))
    assert len(params) == len(space)
    return Trial(params=params)


def tuple_to_trial(data, space):
    """Create a `orion.core.worker.trial.Trial` object from `data`,
    filling only parameter information from `data`.

    :param data: A tuple representing a sample point from `space`.
    :param space: Definition of problem's domain.
    :type space: `orion.algo.space.Space`
    """
    assert len(data) == len(space)
    params = []
    for i, dim in enumerate(space.values()):
        params.append(dict(
            name=dim.name,
            type=dim.type,
            value=data[i]
            ))
    return Trial(params=params)


def get_trial_results(trial):
    """Format results from a `Trial` using standard structures."""
    results = dict()

    lie = trial.lie
    objective = trial.objective

    if lie:
        results['objective'] = lie.value
    elif objective:
        results['objective'] = objective.value
    else:
        results['objective'] = None

    results['constraint'] = [result.value for result in trial.results
                             if result.type == 'constraint']
    grad = trial.gradient
    results['gradient'] = tuple(grad.value) if grad else None

    return results


def standard_param_name(name):
    """Convert parameter name to namespace format"""
    return name.lstrip("/").lstrip("-").replace("-", "_")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
:mod:`orion.core.cli.db.setup` -- Module running the setup command
===============================================================

.. module:: setup
   :platform: Unix
   :synopsis: Creates a configurarion file for the database.
"""

import logging
import os

import yaml

import orion.core
from orion.core.utils.terminal import ask_question

log = logging.getLogger(__name__)
SHORT_DESCRIPTION = 'Starts the database configuration wizard'
DESCRIPTION = """
This command starts the database configuration wizard and creates a configuration file for the
database.
"""


def add_subparser(parser):
    """Return the parser that needs to be used for this command"""
    setup_parser = parser.add_parser('setup', help=SHORT_DESCRIPTION, description=DESCRIPTION)

    setup_parser.set_defaults(func=main)

    return setup_parser


# pylint: disable = unused-argument
def main(*args):
    """Build a configuration file."""
    default_file = orion.core.DEF_CONFIG_FILES_PATHS[-1]

    if os.path.exists(default_file):
        cancel = ''
        while cancel.strip().lower() not in ['y', 'n']:
            cancel = ask_question(
                "This will overwrite {}, do you want to proceed? (y/n) ".format(default_file), "n")

        if cancel.strip().lower() == 'n':
            return

    _type = ask_question("Enter the database type: ", "mongodb")
    name = ask_question("Enter the database name: ", "test")
    host = ask_question("Enter the database host: ", "localhost")

    config = {'database': {'type': _type, 'name': name, 'host': host}}

    print("Default configuration file will be saved at: ")
    print(default_file)

    dirs = '/'.join(default_file.split('/')[:-1])
    os.makedirs(dirs, exist_ok=True)

    with open(default_file, 'w') as output:
        yaml.dump(config, output, default_flow_style=False)



# -*- coding: utf-8 -*-
"""
:mod:`orion.core.utils.module_import` -- Utility functions for importing modules
================================================================================

.. module:: format_trials
   :platform: Unix
   :synopsis: Conversion functions between various data types used in
      framework's ecosystem.

"""

import os


def load_modules_in_path(path, filter_function=None):
    """
    Load all modules inside `path` and return a list of those
    fitting the filter function.
    """
    this_module = __import__(path, fromlist=[''])
    file_path = this_module.__path__[0]

    files = list(map(lambda f: f.split('.')[0],
                     filter(lambda f2: f2.endswith('py'), os.listdir(file_path))))

    modules = map(lambda f: __import__(path + '.' + f, fromlist=['']), files)

    if filter_function is not None:
        modules = filter(filter_function, modules)

    return list(modules)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
:mod:`orion.core.utils.backward` -- Backward compatibility utils
================================================================

.. module:: info
   :platform: Unix
   :synopsis: Helper functions to support backward compatibility

"""

import orion.core
from orion.core.io.orion_cmdline_parser import OrionCmdlineParser


def update_user_args(metadata):
    """Make sure user script is not removed from metadata"""
    if "user_script" in metadata and metadata["user_script"] not in metadata["user_args"]:
        metadata["user_args"] = [metadata["user_script"]] + metadata["user_args"]


def populate_priors(metadata):
    """Compute parser state and priors based on user_args and populate metadata."""
    if 'user_args' not in metadata:
        return

    update_user_args(metadata)

    parser = OrionCmdlineParser(orion.core.config.worker.user_script_config,
                                allow_non_existing_files=True)
    parser.parse(metadata["user_args"])
    metadata["parser"] = parser.get_state_dict()
    metadata["priors"] = dict(parser.priors)


def populate_space(config, force_update=True):
    """Add the space definition at the root of config."""
    if 'space' in config and not force_update:
        return

    populate_priors(config['metadata'])
    # Overwrite space to make sure to include changes from user_args
    if 'priors' in config['metadata']:
        config['space'] = config['metadata']['priors']


def db_is_outdated(database):
    """Return True if the database scheme is outdated."""
    deprecated_indices = [('name', 'metadata.user'), ('name', 'metadata.user', 'version'),
                          'name_1_metadata.user_1', 'name_1_metadata.user_1_version_1']

    index_information = database.index_information('experiments')
    return any(index in deprecated_indices for index in index_information.keys())


def update_db_config(config):
    """Merge DB config back into storage config"""
    config.setdefault('storage', orion.core.config.storage.to_dict())
    if 'database' in config:
        config['storage'] = {'type': 'legacy'}
        config['storage']['database'] = config.pop('database')



# -*- coding: utf-8 -*-
"""
:mod:`orion.storage.base` -- Generic Storage Protocol
=====================================================

.. module:: base
   :platform: Unix
   :synopsis: Implement a generic protocol to allow Orion to communicate using
              different storage backend

"""
import copy
import logging

import orion.core
from orion.core.utils.singleton import AbstractSingletonType, SingletonFactory

log = logging.getLogger(__name__)


def get_uid(item=None, uid=None, force_uid=True):
    """Return uid either from `item` or directly uid.

    Parameters
    ----------
    item: Experiment or Trial, optional
       Object with .id attribute

    uid: str, optional
        str id representation

    force_uid: bool, optional
        If True, at least one of item or uid must be passed.

    Raises
    ------
    UndefinedCall
        if both item and uid are not set and force_uid is True

    AssertionError
        if both item and uid are provided and they do not match
    """
    if item is not None and uid is not None:
        assert item.id == uid

    if uid is None:
        if item is None and force_uid:
            raise MissingArguments('Either `item` or `uid` should be set')
        elif item is not None:
            uid = item.id

    return uid


class FailedUpdate(Exception):
    """Exception raised when we are unable to update a trial' status"""

    pass


class MissingArguments(Exception):
    """Raised when calling a function without the minimal set of parameters"""

    pass


class BaseStorageProtocol(metaclass=AbstractSingletonType):
    """Implement a generic protocol to allow Orion to communicate using
    different storage backend

    """

    def create_experiment(self, config):
        """Insert a new experiment inside the database"""
        raise NotImplementedError()

    def delete_experiment(self, experiment=None, uid=None):
        """Delete matching experiments from the database

        Parameters
        ----------
        experiment: Experiment, optional
           experiment object to retrieve from the database

        uid: str, optional
            experiment id used to retrieve the trial object

        Returns
        -------
        Number of experiments deleted.

        Raises
        ------
        UndefinedCall
            if both experiment and uid are not set

        AssertionError
            if both experiment and uid are provided and they do not match
        """
        raise NotImplementedError()

    def update_experiment(self, experiment=None, uid=None, where=None, **kwargs):
        """Update the fields of a given experiment

        Parameters
        ----------
        experiment: Experiment, optional
           experiment object to retrieve from the database

        uid: str, optional
            experiment id used to retrieve the trial object

        where: Optional[dict]
            constraint experiment must respect

        **kwargs: dict
            a dictionary of fields to update

        Returns
        -------
        returns true if the underlying storage was updated

        Raises
        ------
        UndefinedCall
            if both experiment and uid are not set

        AssertionError
            if both experiment and uid are provided and they do not match

        """
        raise NotImplementedError()

    def fetch_experiments(self, query, selection=None):
        """Fetch all experiments that match the query"""
        raise NotImplementedError()

    def register_trial(self, trial):
        """Create a new trial to be executed"""
        raise NotImplementedError()

    def register_lie(self, trial):
        """Register a *fake* trial created by the strategist.

        The main difference between fake trial and orignal ones is the addition of a fake objective
        result, and status being set to completed. The id of the fake trial is different than the id
        of the original trial, but the original id can be computed using the hashcode on parameters
        of the fake trial. See mod:`orion.core.worker.strategy` for more information and the
        Strategist object and generation of fake trials.

        Parameters
        ----------
        trial: `Trial` object
            Fake trial to register in the database

        """
        raise NotImplementedError()

    def delete_trials(self, experiment=None, uid=None, where=None):
        """Delete matching trials from the database

        Parameters
        ----------
        experiment: Experiment, optional
           experiment object to retrieve from the database

        uid: str, optional
            experiment id used to retrieve the trial object

        where: Optional[dict]
            constraint trials must respect

        Returns
        -------
        Number of trials deleted.

        Raises
        ------
        UndefinedCall
            if both experiment and uid are not set

        AssertionError
            if both experiment and uid are provided and they do not match
        """
        raise NotImplementedError()

    def reserve_trial(self, experiment):
        """Select a pending trial and reserve it for the worker

        Returns
        -------
        Returns the reserved trial or None if no trials were found

        """
        raise NotImplementedError()

    def fetch_trials(self, experiment=None, uid=None):
        """Fetch all the trials of an experiment in the database

        Parameters
        ----------
        experiment: Experiment, optional
           experiment object to retrieve from the database

        uid: str, optional
            experiment id used to retrieve the trial object

        Returns
        -------
        return none if the experiment is not found,

        Raises
        ------
        UndefinedCall
            if both experiment and uid are not set

        AssertionError
            if both experiment and uid are provided and they do not match

        """
        raise NotImplementedError()

    def update_trials(self, experiment=None, uid=None, where=None, **kwargs):
        """Update trials of a given experiment matching a query

        Parameters
        ----------
        experiment: Experiment, optional
           experiment object to retrieve from the database

        uid: str, optional
            experiment id used to retrieve the trial object

        where: Optional[dict]
            constraint trials must respect

        **kwargs: dict
            a dictionary of fields to update

        Raises
        ------
        UndefinedCall
            if both experiment and uid are not set

        AssertionError
            if both experiment and uid are provided and they do not match

        """
        raise NotImplementedError()

    def update_trial(self, trial=None, uid=None, where=None, **kwargs):
        """Update fields of a given trial

        Parameters
        ----------
        trial: Trial, optional
           trial object to update in the database

        uid: str, optional
            id of the trial to update in the database

        where: Optional[dict]
            constraint trials must respect. Note: useful to handle race conditions.

        **kwargs: dict
            a dictionary of fields to update

        Raises
        ------
        UndefinedCall
            if both trial and uid are not set

        AssertionError
            if both trial and uid are provided and they do not match

        """
        raise NotImplementedError()

    def get_trial(self, trial=None, uid=None):
        """Fetch a single trial

        Parameters
        ----------
        trial: Trial, optional
           trial object to retrieve from the database

        uid: str, optional
            trial id used to retrieve the trial object

        Returns
        -------
        return none if the trial is not found,

        Raises
        ------
        UndefinedCall
            if both trial and uid are not set

        AssertionError
            if both trial and uid are provided and they do not match

        """
        raise NotImplementedError()

    def fetch_lost_trials(self, experiment):
        """Fetch all trials that have a heartbeat older than
        some given time delta (2 minutes by default)
        """
        raise NotImplementedError()

    def retrieve_result(self, trial, *args, **kwargs):
        """Fetch the result from a given medium (file, db, socket, etc..) for a given trial and
        insert it into the trial object
        """
        raise NotImplementedError()

    def push_trial_results(self, trial):
        """Push the trial's results to the database"""
        raise NotImplementedError()

    def set_trial_status(self, trial, status, heartbeat=None):
        """Update the trial status and the heartbeat

        Raises
        ------
        FailedUpdate
            The exception is raised if the status of the trial object
            does not match the status in the database

        """
        raise NotImplementedError()

    def fetch_pending_trials(self, experiment):
        """Fetch all trials that are available to be executed by a worker,
        this includes new, suspended and interrupted trials
        """
        raise NotImplementedError()

    def fetch_noncompleted_trials(self, experiment):
        """Fetch all non completed trials"""
        raise NotImplementedError()

    def fetch_trials_by_status(self, experiment, status):
        """Fetch all trials with the given status"""
        raise NotImplementedError()

    def count_completed_trials(self, experiment):
        """Count the number of completed trials"""
        raise NotImplementedError()

    def count_broken_trials(self, experiment):
        """Count the number of broken trials"""
        raise NotImplementedError()

    def update_heartbeat(self, trial):
        """Update trial's heartbeat"""
        raise NotImplementedError()


# pylint: disable=too-few-public-methods,abstract-method
class Storage(BaseStorageProtocol, metaclass=SingletonFactory):
    """Storage protocol is a generic way of allowing Orion to interface with different storage.
    MongoDB, track, cometML, MLFLow, etc...

    Examples
    --------
    >>> Storage('track', uri='file://orion_test.json')
    >>> Storage('legacy', experiment=...)

    Notes
    -----
    When retrieving an already initialized Storage object you should use `get_storage`.
    `Storage()` should only be used for initialization purposes as `get_storage`
    raises more granular error messages.

    """

    pass


def get_storage():
    """Return current storage

    This is a wrapper around the Storage Singleton object to provide
    better error message when it is used without being initialized.

    Raises
    ------
    RuntimeError
        If the underlying storage was not initialized prior to calling this function

    Notes
    -----
    To initialize the underlying storage you must first call `Storage(...)`
    with the appropriate arguments for the chosen backend

    """
    return Storage()


def setup_storage(storage=None, debug=False):
    """Create the storage instance from a configuration.

    Parameters
    ----------
    config: dict, optional
        Configuration for the storage backend. If not defined, global configuration
        is used.
    debug: bool, optional
        If using in debug mode, the storage config is overrided with legacy:EphemeralDB.
        Defaults to False.

    """
    if storage is None:
        storage = orion.core.config.storage.to_dict()

    storage = copy.deepcopy(storage)

    if storage.get('type') == 'legacy' and 'database' not in storage:
        storage['database'] = orion.core.config.storage.database.to_dict()
    elif storage.get('type') is None and 'database' in storage:
        storage['type'] = 'legacy'

    if debug:
        storage = {'type': 'legacy', 'database': {'type': 'EphemeralDB'}}

    storage_type = storage.pop('type')

    log.debug("Creating %s storage client with args: %s", storage_type, storage)
    try:
        Storage(of_type=storage_type, **storage)
    except ValueError:
        if Storage().__class__.__name__.lower() != storage_type.lower():
            raise


# pylint: disable=too-few-public-methods
class ReadOnlyStorageProtocol(object):
    """Read-only interface from a storage protocol.

    .. seealso::

        :py:class:`orion.core.storage.BaseStorageProtocol`
    """

    __slots__ = ('_storage', )
    valid_attributes = {
        'get_trial',
        'fetch_trials',
        'fetch_experiments',
        'count_broken_trials',
        'count_completed_trials',
        'fetch_noncompleted_trials',
        'fetch_pending_trials',
        'fetch_lost_trials',
        'fetch_trials_by_status'
    }

    def __init__(self, protocol):
        """Init method, see attributes of :class:`BaseStorageProtocol`."""
        self._storage = protocol

    def __getattr__(self, attr):
        """Get attribute only if valid"""
        if attr not in self.valid_attributes:
            raise AttributeError("Cannot access attribute %s on ReadOnlyStorageProtocol." % attr)

        return getattr(self._storage, attr)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Simple example to fill db with python api"""
from orion.client import create_experiment


create_experiment(
    'hunt-python',
    space={'x': 'uniform(-50,50)'},
    algorithms={'random': {'seed': 1}})

create_experiment(
    'hunt-python-branch-old',
    space={'x': 'uniform(-50,50)'},
    algorithms={'random': {'seed': 1}},
    branching={'branch_from': 'hunt-python'})



# -*- coding: utf-8 -*-
"""
:mod:`orion.core.resolve_config` -- Configuration parsing and resolving
=======================================================================

.. module:: resolve_config
   :platform: Unix
   :synopsis: How does orion resolve configuration settings?

How:

 - Experiment name resolves like this:
    * cmd-arg **>** cmd-provided orion_config **>** REQUIRED (no default is given)

 - Database options resolve with the following precedence (high to low):
    * cmd-provided orion_config **>** env vars **>** default files **>** defaults

.. seealso:: :const:`ENV_VARS`, :const:`ENV_VARS_DB`


 - All other managerial, `Optimization` or `Dynamic` options resolve like this:

    * cmd-args **>** cmd-provided orion_config **>** database (if experiment name
      can be found) **>** default files

Default files are given as a list at :const:`orion.core.DEF_CONFIG_FILES_PATHS` and a
precedence is respected when building the settings dictionary:

 * default orion example file **<** system-wide config **<** user-wide config

.. note:: `Optimization` entries are required, `Dynamic` entry is optional.

"""
import copy
import getpass
import hashlib
import logging
import os

import git
from numpy import inf as infinity
import yaml

import orion
import orion.core
from orion.core import config
from orion.core.io.orion_cmdline_parser import OrionCmdlineParser
from orion.core.utils.flatten import unflatten


def is_exe(path):
    """Test whether `path` describes an executable file."""
    return os.path.isfile(path) and os.access(path, os.X_OK)


log = logging.getLogger(__name__)

################################################################################
#                 Default Settings and Environmental Variables                 #
################################################################################

# Default settings for command line arguments (option, description)
DEF_CMD_MAX_TRIALS = (infinity, 'inf/until preempted')
DEF_CMD_WORKER_TRIALS = (infinity, 'inf/until preempted')
DEF_CMD_POOL_SIZE = (1, str(1))

# list containing tuples of
# (environmental variable names, configuration keys, default values)
ENV_VARS_DB = [
    ('ORION_DB_NAME', 'name'),
    ('ORION_DB_TYPE', 'type'),
    ('ORION_DB_ADDRESS', 'host'),
    ('ORION_DB_PORT', 'port')
    ]

# TODO: Default resource from environmental (localhost)

# dictionary describing lists of environmental tuples (e.g. `ENV_VARS_DB`)
# by a 'key' to be used in the experiment's configuration dict
ENV_VARS = dict(
    database=ENV_VARS_DB
    )


def _convert_dashes(config, ref):
    """Convert dash in keys to underscores based on a reference dict.

    The reference is used to avoid converting keys in dictionary that are values
    of options.
    """
    config = copy.deepcopy(config)
    for key in list(config.keys()):
        converted_key = key.replace('-', '_')
        if converted_key in ref:
            config[converted_key] = config.pop(key)

            if all(isinstance(item[converted_key], dict) for item in [config, ref]):
                config[converted_key] = _convert_dashes(config[converted_key], ref[converted_key])

    return config


# NOTE: Silencing this pylint error for now, but seriously this function is quite horrible.
#       We'll need to clean this up at some point...
# pylint:disable=too-many-branches
def fetch_config_from_cmdargs(cmdargs):
    """Turn flat cmdargs into nested dicts like orion.core.config."""
    config_file = cmdargs.pop('config', None)
    tmp_cmdargs = copy.deepcopy(cmdargs)
    tmp_cmdargs['config'] = config_file
    cmdargs['config'] = config_file
    cmdargs = tmp_cmdargs

    cmdargs_config = {}

    if cmdargs.get('max_trials') is not None:
        log.warning(
            '--max-trials is deprecated and will be removed in v0.3. '
            'Use --exp-max-trials instead')
        cmdargs_config['experiment.max_trials'] = cmdargs.pop('max_trials')

    if cmdargs.get('worker_trials') is not None:
        log.warning(
            '--worker-trials is deprecated and will be removed in v0.3. '
            'Use --worker-max-trials instead')
        cmdargs_config['worker.max_trials'] = cmdargs.pop('worker_trials')

    mappings = dict(
        experiment=dict(
            exp_max_broken='max_broken',
            exp_max_trials='max_trials'),
        worker=dict(
            worker_max_broken='max_broken',
            worker_max_trials='max_trials'))

    mappings = dict(
        experiment=dict(
            max_broken='exp_max_broken',
            max_trials='exp_max_trials'),
        worker=dict(
            max_broken='worker_max_broken',
            max_trials='worker_max_trials'))

    global_config = config.to_dict()

    for key in ['config', 'user_args']:
        if cmdargs.get(key) not in [False, None]:
            cmdargs_config[key] = cmdargs[key]

    for key in ['name', 'user', 'version']:
        if cmdargs.get(key) not in [False, None]:
            cmdargs_config[f'experiment.{key}'] = cmdargs[key]

    for key in ['branch_from', 'branch_to']:
        if cmdargs.get(key) not in [False, None]:
            cmdargs_config[f'evc.{key}'] = cmdargs[key]

    # Apply config at the root
    for key in ['debug']:

        # Adapt to cli arguments
        cli_key = mappings.get(key, key)

        value = cmdargs.pop(cli_key, None)
        if value is not None:
            cmdargs_config[f'{key}'] = value

    # Apply to subconfigs
    for key in ['experiment', 'worker', 'evc']:
        for subkey in global_config[key].keys():

            # Adapt to cli arguments
            cli_key = mappings.get(key, {}).get(subkey, subkey)

            value = cmdargs.pop(cli_key, None)
            if value is not None:
                cmdargs_config[f'{key}.{subkey}'] = value

    return unflatten(cmdargs_config)


def fetch_config(args):
    """Return the config inside the .yaml file if present."""
    orion_file = args.get('config')
    local_config = {}
    if orion_file:
        log.debug("Found orion configuration file at: %s", os.path.abspath(orion_file.name))
        orion_file.seek(0)
        tmp_config = yaml.safe_load(orion_file)

        global_config = config.to_dict()

        tmp_config = _convert_dashes(tmp_config, global_config)

        # Fix deprecations first because some names are shared by experiment and worker
        max_trials = tmp_config.pop('max_trials', None)
        if max_trials is not None:
            log.warning(
                '(DEPRECATED) Option `max_trials` is deprecated '
                'and will be removed in v0.3. Use instead the option'
                '\nexperiment:\n  max_trials: %s', max_trials)
            local_config['experiment.max_trials'] = max_trials

        worker_trials = tmp_config.get('experiment', {}).pop('worker_trials', None)
        if worker_trials is not None:
            log.warning(
                '(DEPRECATED) Option `experiment.worker_trials` is deprecated '
                'and will be removed in v0.3. Use instead the option'
                '\nworker:\n  max_trials: %s', worker_trials)
            local_config['worker.max_trials'] = worker_trials

        worker_trials = tmp_config.pop('worker_trials', None)
        if worker_trials is not None:
            log.warning(
                '(DEPRECATED) Option `worker_trials` is deprecated '
                'and will be removed in v0.3. Use instead the option'
                '\nworker:\n  max_trials: %s', worker_trials)
            local_config['worker.max_trials'] = worker_trials

        producer = tmp_config.pop('producer', None)
        if producer is not None:
            log.warning(
                '(DEPRECATED) Option `producer` is deprecated '
                'and will be removed in v0.3. Use instead the option'
                '\nexperiment:\n  strategy: %s', producer['strategy'])
            local_config['experiment.strategy'] = producer['strategy']

        producer = tmp_config.get('experiment', {}).pop('producer', None)
        if producer is not None:
            log.warning(
                '(DEPRECATED) Option `experiment.producer` is deprecated '
                'and will be removed in v0.3. Use instead the option'
                '\nexperiment:\n  strategy: %s', producer['strategy'])
            local_config['experiment.strategy'] = producer['strategy']

        local_config = unflatten(local_config)

        # For backward compatibility
        for key in ['storage', 'experiment', 'worker', 'evc']:
            subkeys = list(global_config[key].keys())

            # Arguments that are only supported locally
            if key == 'experiment':
                subkeys += ['name', 'version', 'user']
            elif key == 'evc':
                subkeys += ['branch_from', 'branch_to']

            for subkey in subkeys:
                # Backward compatibility
                backward_value = tmp_config.pop(subkey, None)
                if backward_value is not None:
                    log.warning(
                        '(DEPRECATED) Option `%s` and will be removed in v0.3. '
                        'Use instead the option' '\n%s:\n  %s:\n    %s',
                        subkey, key, subkey, yaml.dump(backward_value, indent=6))
                value = tmp_config.get(key, {}).pop(subkey, backward_value)
                if value is not None:
                    local_config.setdefault(key, {})
                    local_config[key][subkey] = value

    return local_config


def fetch_default_options():
    """Create a dict with options from the default configuration files.

    Respect precedence from application's default, to system's and
    user's.

    .. seealso:: :const:`orion.core.DEF_CONFIG_FILES_PATHS`

    """
    default_config = dict()

    # get some defaults
    default_config['name'] = None
    default_config['user'] = getpass.getuser()
    default_config['max_trials'] = DEF_CMD_MAX_TRIALS[0]
    default_config['worker_trials'] = DEF_CMD_WORKER_TRIALS[0]
    default_config['pool_size'] = DEF_CMD_POOL_SIZE[0]
    default_config['algorithms'] = 'random'

    # get default options for some managerial variables (see :const:`ENV_VARS`)
    for signifier, env_vars in ENV_VARS.items():
        default_config[signifier] = {}
        for _, key in env_vars:
            default_config[signifier][key] = config[signifier][key]

    # fetch options from default configuration files
    for configpath in orion.core.DEF_CONFIG_FILES_PATHS:
        try:
            with open(configpath) as f:
                cfg = yaml.safe_load(f)
                if cfg is None:
                    continue
                # implies that yaml must be in dict form
                for k, v in cfg.items():
                    if k in ENV_VARS:
                        default_config[k] = {}
                        for vk, vv in v.items():
                            default_config[k][vk] = vv
                    else:
                        if k != 'name':
                            default_config[k] = v
        except IOError as e:  # default file could not be found
            log.debug(e)
        except AttributeError as e:
            log.warning("Problem parsing file: %s", configpath)
            log.warning(e)

    return default_config


def fetch_env_vars():
    """Fetch environmental variables related to orion's managerial data."""
    env_vars = {}

    for signif, evars in ENV_VARS.items():
        env_vars[signif] = {}

        for var_name, key in evars:
            value = os.getenv(var_name)

            if value is not None:
                env_vars[signif][key] = value

    return env_vars


def fetch_metadata(user=None, user_args=None):
    """Infer rest information about the process + versioning"""
    metadata = {'user': user if user else getpass.getuser()}

    metadata['orion_version'] = orion.core.__version__

    if user_args is None:
        user_args = []

    # Trailing white space are catched by argparse as an empty argument
    if len(user_args) == 1 and user_args[0] == '':
        user_args = []

    cmdline_parser = OrionCmdlineParser(config.worker.user_script_config)
    cmdline_parser.parse(user_args)

    if cmdline_parser.user_script:
        # TODO: Remove this, it is all in cmdline_parser now
        metadata['user_script'] = cmdline_parser.user_script
        metadata['VCS'] = infer_versioning_metadata(cmdline_parser.user_script)

    if user_args:
        # TODO: Remove this, it is all in cmdline_parser now
        metadata['user_args'] = user_args

    return metadata


def merge_configs(*configs):
    """Merge configuration dictionaries following the given hierarchy

    Suppose function is called as merge_configs(A, B, C). Then any pair (key, value) in C would
    overwrite any previous value from A or B. Same apply for B over A.

    If for some pair (key, value), the value is a dictionary, then it will either overwrite previous
    value if it was not also a directory, or it will be merged following
    `merge_configs(old_value, new_value)`.

    .. warning:

        Redefinition of subdictionaries may lead to confusing results because merges do not remove
        data.

        If for instance, we have {'a': {'b': 1, 'c': 2}} and we would like to update `'a'` such that
        it only have `{'c': 3}`, it won't work with {'a': {'c': 3}}.

        merge_configs({'a': {'b': 1, 'c': 2}}, {'a': {'c': 3}}) -> {'a': {'b': 1, 'c': 3}}

    Examples
    --------
    .. code-block:: python
        :linenos:

        a = {'a': 1, 'b': {'c': 2}}
        b = {'b': {'c': 3}}
        c = {'b': {'c': {'d': 4}}}

        m = resolve_config.merge_configs(a, b, c)

        assert m == {'a': 1, 'b': {'c': {'d': 4}}}

        a = {'a': 1, 'b': {'c': 2, 'd': 3}}
        b = {'b': {'c': 4}}
        c = {'b': {'c': {'e': 5}}}

        m = resolve_config.merge_configs(a, b, c)

        assert m == {'a': 1, 'b': {'c': {'e': 5}, 'd': 3}}

    """
    merged_config = configs[0]

    for config_i in configs[1:]:
        for key, value in config_i.items():
            if isinstance(value, dict) and isinstance(merged_config.get(key), dict):
                merged_config[key] = merge_configs(merged_config[key], value)
            elif value is not None:
                merged_config[key] = value

    return merged_config


def fetch_user_repo(user_script):
    """Fetch the GIT repo and its root path given user's script."""
    dir_path = os.path.dirname(os.path.abspath(user_script))
    try:
        git_repo = git.Repo(dir_path, search_parent_directories=True)
    except git.exc.InvalidGitRepositoryError:
        git_repo = None
        logging.warning('Script %s is not in a git repository. Code modification '
                        'won\'t be detected.', os.path.abspath(user_script))
    return git_repo


def infer_versioning_metadata(user_script):
    """
    Infer information about user's script versioning if available.
    Fills the following information in VCS:

    `is_dirty` shows whether the git repo is at a clean state.
    `HEAD_sha` gives the hash of head of the repo.
    `active_branch` shows the active branch of the repo.
    `diff_sha` shows the hash of the diff in the repo.

    :returns: the `VCS` but filled with above info.

    """
    git_repo = fetch_user_repo(user_script)
    if not git_repo:
        return {}
    vcs = {}
    vcs['type'] = 'git'
    vcs['is_dirty'] = git_repo.is_dirty()
    vcs['HEAD_sha'] = git_repo.head.object.hexsha
    if git_repo.head.is_detached:
        vcs['active_branch'] = None
    else:
        vcs['active_branch'] = git_repo.active_branch.name
    # The 'diff' of the current version from the latest commit
    diff = git_repo.git.diff(git_repo.head.commit.tree).encode('utf-8')
    diff_sha = hashlib.sha256(diff).hexdigest()
    vcs['diff_sha'] = diff_sha
    return vcs



# -*- coding: utf-8 -*-
"""
:mod:`orion.core.worker.trials_history` -- History of past trials
=================================================================

.. module:: trials_history
    :platform: Unix
    :synopsis: Maintain the history of past trials used by an algorithm

"""


# pylint:disable=protected-access,too-few-public-methods
class TrialsHistory:
    """Maintain a list of all the last seen trials that are on different dependency paths"""

    def __init__(self):
        """Create empty trials history"""
        self.children = []
        self.ids = set()

    def __contains__(self, trial):
        """Return True if the trial is in the observed history"""
        return trial.id in self.ids

    def update(self, trials):
        """Update the list of children trials

        The children history only keeps children. Current children that are now ancestors of
        the new nodes are discarded from the history. This is because we can rebuild the entire
        history from the current children, therefore we only need to keep those.
        """
        descendents = set(self.children)
        for trial in trials:
            descendents -= set(trial.parents)
            descendents.add(trial.id)

        self.ids |= descendents

        self.children = list(sorted(descendents))
